# source: https://medium.com/cuddle-ai/async-architecture-with-fastapi-celery-and-rabbitmq-c7d029030377
from celery import Celery
from celery.result import AsyncResult

from lexy.core.config import settings as lexy_settings
from lexy.core.celery_config import settings as celery_settings


def create_celery(celery_settings=celery_settings, lexy_settings=lexy_settings):
    celery_app = Celery()
    celery_app.config_from_object(celery_settings, namespace="CELERY")

    # list of modules to import when the Celery worker starts.
    # celery_app.conf.update(imports=(
    #     'myapp.tasks',
    # ))
    # TODO: why is 'lexy.core.celery_tasks' not included in imports?
    celery_app.conf.update(imports=list(lexy_settings.worker_transformer_imports))
    celery_app.conf.update(include=lexy_settings.worker_transformer_imports)

    return celery_app


def get_task_info(task_id, verbose: bool = False) -> dict:
    """Return task info for the given task_id."""
    task_result = AsyncResult(task_id)
    result = {
        "id": task_id,
        "name": task_result.name,
        "status": task_result.status,
        "date_done": task_result.date_done,
        "retries": task_result.retries,
        "worker": task_result.worker,
        "queue": task_result.queue,
        "parent": task_result.parent,
        "children": task_result.children,
    }
    if verbose:
        result["result"] = task_result.result
        result["traceback"] = task_result.traceback
        result["args"] = task_result.args
        result["kwargs"] = task_result.kwargs
    return result



# reference: https://docs.celeryq.dev/en/stable/userguide/testing.html


class TestCelery:
    def test_celery_app(self, celery_app, celery_worker, settings, celery_settings):
        print(f"{celery_app = }")
        print(f"{celery_app.conf.broker_url = }")
        print(f"{celery_app.conf.result_backend = }")
        assert celery_app

        # assert that result_backend is set to the value in test settings
        assert celery_app.conf.result_backend == celery_settings.result_backend

        assert "lexy.transformers.text.embeddings.minilm" in celery_app.tasks
        assert "lexy.db.save_records_to_index" in celery_app.tasks

        save_records_task = celery_app.tasks.get("lexy.db.save_records_to_index")
        assert save_records_task is not None
        assert save_records_task.db.bind.engine.url.database == "lexy_tests"
        assert (
            save_records_task.db.bind.engine.url.render_as_string(hide_password=False)
            == settings.sync_database_url
        )

        text_embeddings_task = celery_app.tasks.get(
            "lexy.transformers.text.embeddings.minilm"
        )
        assert text_embeddings_task is not None

        eager_task = text_embeddings_task.apply(args=["hello world"])
        eager_result = eager_task.get().tolist()
        assert isinstance(eager_result, list)
        assert all(isinstance(elem, float) for elem in eager_result)

        eager_task = text_embeddings_task.apply(args=[["hello", "world"]])
        eager_result = eager_task.get().tolist()
        assert isinstance(eager_result, list)
        assert all(isinstance(elem, list) for elem in eager_result)
        assert all(isinstance(elem, float) for elem in eager_result[0])
        assert all(isinstance(elem, float) for elem in eager_result[1])

        # this hangs unless we include the celery_worker fixture
        task = text_embeddings_task.apply_async(args=[["hello", "world"]])
        assert isinstance(task.id, str)
        result = task.get().tolist()
        assert isinstance(result, list)
        assert all(isinstance(elem, list) for elem in result)
        assert all(isinstance(elem, float) for elem in result[0])
        assert all(isinstance(elem, float) for elem in result[1])

        # this times out unless we include the celery_worker fixture
        task2 = text_embeddings_task.delay("hello world")
        assert isinstance(task2.id, str)
        result2 = task2.get(timeout=10).tolist()
        assert isinstance(result2, list)
        assert all(isinstance(elem, float) for elem in result2)

    def test_celery_config(self, celery_config):
        assert celery_config

    def test_celery_worker(self, celery_worker):
        print(f"{celery_worker = }")
        assert celery_worker

    def test_celery_session_worker(self, celery_session_worker):
        print(f"{celery_session_worker = }")
        assert celery_session_worker



from fastapi import APIRouter, Depends, HTTPException, status
from sqlmodel import select
from sqlmodel.ext.asyncio.session import AsyncSession

from lexy.core.celery_tasks import celery, convert_arrays_to_lists
from lexy.db.session import get_session
from lexy.models.document import DocumentCreate
from lexy.models.transformer import (
    Transformer,
    TransformerCreate,
    TransformerUpdate,
    default_celery_task_name,
)


router = APIRouter()


@router.get(
    "/transformers",
    response_model=list[Transformer],
    status_code=status.HTTP_200_OK,
    name="get_transformers",
    description="Get all transformers",
)
async def get_transformers(
    session: AsyncSession = Depends(get_session),
) -> list[Transformer]:
    result = await session.exec(select(Transformer))
    transformers = result.all()
    return transformers


@router.post(
    "/transformers",
    response_model=Transformer,
    status_code=status.HTTP_201_CREATED,
    name="create_transformer",
    description="Create a transformer",
)
async def create_transformer(
    transformer: TransformerCreate, session: AsyncSession = Depends(get_session)
) -> Transformer:
    # Check if transformer already exists
    existing_transformer = await session.get(Transformer, transformer.transformer_id)
    if existing_transformer:
        raise HTTPException(
            status_code=400, detail="Transformer with that ID already exists"
        )
    db_transformer = Transformer.model_validate(transformer)
    # Verify that the Celery task exists
    celery_task_name = default_celery_task_name(db_transformer.transformer_id)
    registered_tasks = celery.control.inspect().registered_tasks()
    celery_tasks = set().union(*registered_tasks.values())
    if celery_task_name not in celery_tasks:
        raise HTTPException(
            status_code=400, detail=f"Could not find Celery task '{celery_task_name}'"
        )
    session.add(db_transformer)
    await session.commit()
    await session.refresh(db_transformer)
    return db_transformer


@router.get(
    "/transformers/{transformer_id}",
    response_model=Transformer,
    status_code=status.HTTP_200_OK,
    name="get_transformer",
    description="Get a transformer",
)
async def get_transformer(
    transformer_id: str, session: AsyncSession = Depends(get_session)
) -> Transformer:
    result = await session.exec(
        select(Transformer).where(Transformer.transformer_id == transformer_id)
    )
    transformer = result.first()
    if not transformer:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Transformer not found"
        )
    return transformer


@router.patch(
    "/transformers/{transformer_id}",
    response_model=Transformer,
    status_code=status.HTTP_200_OK,
    name="update_transformer",
    description="Update a transformer",
)
async def update_transformer(
    transformer_id: str,
    transformer: TransformerUpdate,
    session: AsyncSession = Depends(get_session),
) -> Transformer:
    result = await session.exec(
        select(Transformer).where(Transformer.transformer_id == transformer_id)
    )
    db_transformer = result.first()
    if not db_transformer:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Transformer not found"
        )
    transformer_data = transformer.model_dump(exclude_unset=True)
    for key, value in transformer_data.items():
        setattr(db_transformer, key, value)
    # Verify that the Celery task exists
    celery_task_name = default_celery_task_name(db_transformer.transformer_id)
    registered_tasks = celery.control.inspect().registered_tasks()
    celery_tasks = set().union(*registered_tasks.values())
    if celery_task_name not in celery_tasks:
        raise HTTPException(
            status_code=400, detail=f"Could not find Celery task '{celery_task_name}'"
        )
    session.add(db_transformer)
    await session.commit()
    await session.refresh(db_transformer)
    return db_transformer


@router.delete(
    "/transformers/{transformer_id}",
    status_code=status.HTTP_200_OK,
    name="delete_transformer",
    description="Delete a transformer",
)
async def delete_transformer(
    transformer_id: str, session: AsyncSession = Depends(get_session)
) -> dict:
    result = await session.exec(
        select(Transformer).where(Transformer.transformer_id == transformer_id)
    )
    transformer = result.first()
    if not transformer:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Transformer not found"
        )
    await session.delete(transformer)
    await session.commit()
    return {"msg": "Transformer deleted", "transformer_id": transformer_id}


@router.post(
    "/transformers/{transformer_id}",
    response_model=dict,
    status_code=status.HTTP_200_OK,
    name="transform_document",
    description="Transform a document",
)
async def transform_document(
    transformer_id: str,
    document: DocumentCreate,
    celery_task_name: str = None,
    transformer_params: dict = None,
    content_only: bool = False,
    session: AsyncSession = Depends(get_session),
) -> dict:
    # TODO: set up a different end point for running a Celery task without needing to
    #  pass a transformer_id
    if celery_task_name:
        task_name = celery_task_name
    else:
        result = await session.exec(
            select(Transformer).where(Transformer.transformer_id == transformer_id)
        )
        transformer = result.first()
        if not transformer:
            raise HTTPException(
                status_code=status.HTTP_404_NOT_FOUND, detail="Transformer not found"
            )
        task_name = transformer.celery_task_name

    if content_only:
        task_arg = document.content
    else:
        task_arg = document

    task = celery.send_task(
        task_name,
        args=[task_arg],
        kwargs=transformer_params,
    )
    result = task.get()
    response = {"task_id": task.id, "result": result}
    return convert_arrays_to_lists(response)



from datetime import datetime
from typing import Any, Optional
from uuid import uuid4, UUID

from sqlalchemy import DateTime, func
from sqlalchemy.dialects.postgresql import JSONB
from sqlalchemy.sql.schema import Table
from sqlmodel import SQLModel, Field


# Fields requiring SQLAlchemy objects (Column or ForeignKey) have been moved to
#   IndexManager until SQLModel fixes inheritance for Pydantic V2+


class IndexRecordBase(SQLModel):
    document_id: Optional[UUID] = Field(
        foreign_key="documents.document_id", index=True, nullable=True
    )
    custom_id: Optional[str] = Field(default=None)
    meta: Optional[dict[Any, Any]] = Field(default={}, sa_type=JSONB)


class IndexRecordCreate(IndexRecordBase):
    pass


class IndexRecordUpdate(IndexRecordBase):
    custom_id: Optional[str] = None
    meta: Optional[dict[Any, Any]] = None


class IndexRecordBaseTable(IndexRecordBase):
    __table__: Table | None = None
    index_record_id: UUID = Field(
        default_factory=uuid4,
        primary_key=True,
        index=True,
        nullable=False,
    )
    # `created_at` and `updated_at` are hacked to use sa_type along with kwargs -- will
    #  update once SQLModel inheritance is fixed
    created_at: datetime = Field(
        default=None,
        sa_type=DateTime(timezone=True),  # type: ignore
        sa_column_kwargs=dict(nullable=False, server_default=func.now()),
    )
    updated_at: datetime = Field(
        default=None,
        sa_type=DateTime(timezone=True),  # type: ignore
        sa_column_kwargs=dict(
            nullable=False, server_default=func.now(), onupdate=func.now()
        ),
    )
    task_id: Optional[UUID] = Field(default=None)
    binding_id: Optional[int] = Field(
        foreign_key="bindings.binding_id", index=True, nullable=True
    )



import logging
import time

from fastapi import APIRouter, Depends, HTTPException, status
from sqlmodel import select
from sqlmodel.ext.asyncio.session import AsyncSession

from lexy.db.session import get_session
from lexy.models.index import Index, IndexCreate, IndexUpdate
from lexy.core.events import restart_celery_worker
from lexy.api.deps import index_manager


logger = logging.getLogger(__name__)
router = APIRouter()

# TODO: move this to the proper config location
celery_db_worker = "celery@celeryworker"


@router.get(
    "/indexes",
    response_model=list[Index],
    status_code=status.HTTP_200_OK,
    name="get_indexes",
    description="Get all indexes",
)
async def get_indexes(session: AsyncSession = Depends(get_session)) -> list[Index]:
    result = await session.exec(select(Index))
    indexes = result.all()
    return indexes


@router.post(
    "/indexes",
    status_code=status.HTTP_201_CREATED,
    name="create_index",
    description="Create a new index",
)
async def create_index(
    index: IndexCreate, session: AsyncSession = Depends(get_session)
) -> Index:
    # check if index already exists
    existing_index = await session.get(Index, index.index_id)
    if existing_index:
        raise HTTPException(status_code=400, detail="Index already exists")

    # create new index
    db_index = Index(**index.model_dump())
    session.add(db_index)
    await session.commit()
    await session.refresh(db_index)

    # create the index model and table
    _ = index_manager.create_index_model_and_table(index_id=db_index.index_id)
    logger.info(
        f"Restarting db worker '{celery_db_worker}' following creation of index table "
        f"for index_id: {db_index.index_id}"
    )
    time.sleep(0.5)
    celery_restart_response = restart_celery_worker(celery_db_worker)
    logger.info(f"{celery_restart_response = }")

    return db_index


@router.get(
    "/indexes/{index_id}",
    response_model=Index,
    status_code=status.HTTP_200_OK,
    name="get_index",
    description="Get an index",
)
async def get_index(
    index_id: str, session: AsyncSession = Depends(get_session)
) -> Index:
    result = await session.exec(select(Index).where(Index.index_id == index_id))
    index = result.first()
    if not index:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Index not found"
        )
    return index


@router.patch(
    "/indexes/{index_id}",
    response_model=Index,
    status_code=status.HTTP_200_OK,
    name="update_index",
    description="Update an index",
)
async def update_index(
    index_id: str, index: IndexUpdate, session: AsyncSession = Depends(get_session)
) -> Index:
    result = await session.exec(select(Index).where(Index.index_id == index_id))
    db_index = result.first()
    if not db_index:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Index not found"
        )
    index_data = index.model_dump(exclude_unset=True)
    for key, value in index_data.items():
        setattr(db_index, key, value)
    session.add(db_index)
    await session.commit()
    await session.refresh(db_index)
    return db_index


@router.delete(
    "/indexes/{index_id}",
    status_code=status.HTTP_200_OK,
    name="delete_index",
    description="Delete an index",
)
async def delete_index(
    index_id: str,
    drop_table: bool = False,
    session: AsyncSession = Depends(get_session),
) -> dict:
    result = await session.exec(select(Index).where(Index.index_id == index_id))
    index = result.first()
    if not index:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Index not found"
        )

    index_table_name = index.index_table_name

    # NOTE: Deleting the index with option `drop_table=False` will leave the index
    #   model in the index manager. To remove it, recreate the index, and then delete
    #   with `drop_table=True`. The issue with removing it from the index manager is
    #   created when trying to recreate a table with the same name.
    table_dropped = False
    if drop_table:
        table_dropped = index_manager.drop_index_table(index_id=index_id)

    await session.delete(index)
    await session.commit()

    # TODO: restart celery worker here?
    return {
        "msg": "Index deleted",
        "index_id": index_id,
        "index_table_name": index_table_name,
        "table_dropped": table_dropped,
    }



from fastapi import APIRouter

from lexy.api.endpoints import (
    bindings,
    collections,
    documents,
    indexes,
    index_records,
    login,
    transformers,
    users,
    utils,
)


lexy_api = APIRouter()

lexy_api.include_router(bindings.router, tags=["bindings"])
lexy_api.include_router(collections.router, tags=["collections"])
lexy_api.include_router(documents.router, tags=["documents"])
lexy_api.include_router(indexes.router, tags=["indexes"])
lexy_api.include_router(index_records.router, tags=["index_records"])
lexy_api.include_router(login.router, tags=["login"])
lexy_api.include_router(transformers.router, tags=["transformers"])
lexy_api.include_router(users.router, tags=["users"])
lexy_api.include_router(utils.router, tags=["utils"])



from typing import TYPE_CHECKING, Union

from lexy.core.config import settings
from lexy.storage.base import StorageClient
from lexy.storage.gcs import GCSClient, GoogleCredentialsError
from lexy.storage.s3 import S3Client

if TYPE_CHECKING:
    from lexy.models.document import Document, DocumentBase


# TODO: Add caching
# TODO: Add argument for storage service and default to settings.DEFAULT_STORAGE_SERVICE
# TODO: Move this to lexy.api.deps?
async def get_storage_client() -> StorageClient | None:
    if settings.DEFAULT_STORAGE_SERVICE == "s3":
        # The boto3 client allows for initialization without credentials
        s3_client = S3Client()
        yield s3_client
    elif settings.DEFAULT_STORAGE_SERVICE == "gcs":
        # The google-cloud-storage client requires credentials
        try:
            gcs_client = GCSClient()
            yield gcs_client
        except GoogleCredentialsError:
            yield None
    else:
        raise ValueError("Unsupported storage service configured")


def generate_signed_urls_for_document(
    document: Union["Document", "DocumentBase"],
    storage_client: StorageClient,
    expiration: int = 3600,
) -> dict:
    """Generate signed URLs for a document.

    Args:
        document (Document): The document object.
        storage_client (StorageClient): The storage client.
        expiration (int): The number of seconds the presigned URLs are valid for.
            Default is 3600 seconds (1 hour).

    Returns:
        Dict: A dictionary containing presigned URLs.

    Examples:
        >>> from lexy_py import LexyClient
        >>> lx = LexyClient()
        >>> my_image_document = lx.list_documents(collection_name='my_image_collection', limit=1)[0]
        >>> generate_signed_urls_for_document(my_image_document, storage_client=storage_client)
        {
            "object": "https://my-bucket.s3.amazonaws.com/path/to/object?...",
            "thumbnails": {
                "256x256": "https://my-bucket.s3.amazonaws.com/path/to/thumbnail?..."
            }
        }
        >>> my_pdf_document = lx.list_documents(collection_name='pdf_collection', limit=1)[0]
        >>> generate_signed_urls_for_document(my_pdf_document, storage_client=storage_client)
        {
            "object": "https://my-bucket.s3.amazonaws.com/path/to/object?...",
        }
    """
    presigned_urls = {}

    # url for the document object
    if document.meta.get("storage_bucket") and document.meta.get("storage_key"):
        presigned_urls["object"] = storage_client.generate_presigned_url(
            bucket_name=document.meta.get("storage_bucket"),
            object_name=document.meta.get("storage_key"),
            expiration=expiration,
        )

    # urls for thumbnails
    if document.meta.get("image") and document.meta.get("image").get("thumbnails"):
        presigned_urls["thumbnails"] = {}
        for dims, vals in document.meta.get("image").get("thumbnails").items():
            presigned_urls["thumbnails"][dims] = storage_client.generate_presigned_url(
                bucket_name=vals.get("storage_bucket"),
                object_name=vals.get("storage_key"),
                expiration=expiration,
            )

    return presigned_urls


async def construct_key_for_document(
    document: "Document" = None,
    collection_id: str = None,
    document_id: str = None,
    path_prefix: str = None,
    filename: str = None,
) -> str:
    """Construct a storage key (object name) for a document.

    Args:
        document (Document): The document object.
        collection_id (str): The collection ID.
        document_id (str): The document ID.
        path_prefix (str): The path prefix.
        filename (str): The filename.

    Returns:
        str: The storage key (object name).

    Raises:
        ValueError: If neither a document object nor collection_id and document_id are
            provided.

    Examples:
        >>> from lexy_py import LexyClient
        >>> lx = LexyClient()
        >>> my_image_document = lx.list_documents(collection_name='my_image_collection', limit=1)[0]
        >>> construct_key_for_document(my_image_document)
        'collections/423c718b/documents/123456'
        >>> construct_key_for_document(collection_id='423c718b', document_id='123456')
        'collections/423c718b/documents/123456'
        >>> construct_key_for_document(collection_id='423c718b', document_id='123456', filename='image.jpg')
        'collections/423c718b/documents/123456/image.jpg'
        >>> construct_key_for_document(document=my_image_document, path_prefix='public')
        'public/collections/423c718b/documents/123456'
    """
    if document and document.collection_id and document.document_id:
        key = f"collections/{document.collection_id}/documents/{document.document_id}"
    elif collection_id and document_id:
        key = f"collections/{collection_id}/documents/{document_id}"
    else:
        raise ValueError(
            "Either a document object or collection_id and document_id must be "
            "provided."
        )
    if path_prefix:
        key = f"{path_prefix.rstrip('/')}/{key}"
    if filename:
        key = f"{key}/{filename}"
    return key


async def construct_key_for_thumbnail(
    dims: tuple[int, int],
    document: "Document" = None,
    collection_id: str = None,
    document_id: str = None,
    path_prefix: str = None,
    filename: str = None,
) -> str:
    """Construct a storage key (object name) for thumbnails of a document.

    Args:
        dims (tuple[int, int]): The dimensions of the thumbnail.
        document (Document): The document object.
        collection_id (str): The collection ID.
        document_id (str): The document ID.
        path_prefix (str): The path prefix.
        filename (str): The filename.

    Returns:
        str: The storage key (object name).

    Raises:
        ValueError: If neither a document object nor collection_id and document_id are
            provided.

    Examples:
        >>> from lexy_py import LexyClient
        >>> lx = LexyClient()
        >>> my_image_document = lx.list_documents(collection_name='my_image_collection', limit=1)[0]
        >>> construct_key_for_thumbnail(dims=(200, 200), document=my_image_document)
        'collections/423c718b/thumbnails/200x200/123456'
        >>> construct_key_for_thumbnail(dims=(200, 200), collection_id='423c718b', document_id='123456')
        'collections/423c718b/thumbnails/200x200/123456'
        >>> construct_key_for_thumbnail(dims=(150, 150), collection_id='423c718b',
        ...                             document_id='123456', filename='thumbnail.jpg')
        'collections/423c718b/thumbnails/150x150/123456/thumbnail.jpg'
        >>> construct_key_for_thumbnail(dims=(200, 200), document=my_image_document, path_prefix='public')
        'public/collections/423c718b/thumbnails/200x200/123456'
    """
    if not dims:
        raise ValueError("Thumbnail dimensions must be provided.")
    if document and document.collection_id and document.document_id:
        key = f"collections/{document.collection_id}/thumbnails/{dims[0]}x{dims[1]}/{document.document_id}"
    elif collection_id and document_id:
        key = (
            f"collections/{collection_id}/thumbnails/{dims[0]}x{dims[1]}/{document_id}"
        )
    else:
        raise ValueError(
            "Either a document object or collection_id and document_id must be "
            "provided."
        )
    if path_prefix:
        key = f"{path_prefix.rstrip('/')}/{key}"
    if filename:
        key = f"{key}/{filename}"
    return key



import datetime
import os
import time
import warnings

import boto3
import pytest
from botocore.client import Config
from botocore.exceptions import ClientError, NoCredentialsError
from google.cloud import storage
from google.oauth2 import service_account

from lexy.storage import signed_url_is_expired
from lexy.storage.client import construct_key_for_document, construct_key_for_thumbnail
from lexy.storage.gcs import GCSClient
from lexy.storage.s3 import S3Client
from lexy.models.document import Document


@pytest.fixture(scope="module")
def s3():
    s3_client = boto3.client("s3")
    try:
        s3_client.list_buckets()
        yield s3_client
    except NoCredentialsError:
        warnings.warn("S3 credentials are not available", UserWarning)
        pytest.skip("S3 credentials are not available")
    except ClientError as e:
        if e.response["Error"]["Code"] in ("AccessDenied", "403"):
            warnings.warn(f"S3 client access denied: {e.response}", UserWarning)
            pytest.skip("S3 access is denied")
        else:
            warnings.warn(f"S3 client has an error: {e.response}", UserWarning)
            pytest.skip("S3 client has an error")


@pytest.fixture(scope="module")
def s3v4():
    s3v4_client = boto3.client("s3", config=Config(signature_version="s3v4"))
    try:
        s3v4_client.list_buckets()
        yield s3v4_client
    except NoCredentialsError:
        warnings.warn("S3 credentials are not available", UserWarning)
        pytest.skip("S3 credentials are not available")
    except ClientError as e:
        if e.response["Error"]["Code"] in ("AccessDenied", "403"):
            warnings.warn(f"S3 client access denied: {e.response}", UserWarning)
            pytest.skip("S3 access is denied")
        else:
            warnings.warn(f"S3 client has an error: {e.response}", UserWarning)
            pytest.skip("S3 client has an error")


@pytest.fixture(scope="module")
def gcs(settings):
    credentials_file = settings.GOOGLE_APPLICATION_CREDENTIALS
    if not credentials_file:
        warnings.warn("GOOGLE_APPLICATION_CREDENTIALS is not set", UserWarning)
        pytest.skip("GOOGLE_APPLICATION_CREDENTIALS is not set")

    print(f"Creating GCS client using credentials file: {credentials_file}")
    credentials = service_account.Credentials.from_service_account_file(
        credentials_file
    )
    yield storage.Client(credentials=credentials)


@pytest.fixture(scope="module")
def lx_s3():
    s3_client = S3Client()
    if not s3_client.is_authenticated():
        warnings.warn("S3 client is not authenticated", UserWarning)
        pytest.skip("S3 client is not authenticated")
    yield s3_client


@pytest.fixture(scope="module")
def lx_s3v4():
    config = Config(signature_version="s3v4")
    s3_client = S3Client(config=config)
    if not s3_client.is_authenticated():
        warnings.warn("S3 client is not authenticated", UserWarning)
        pytest.skip("S3 client is not authenticated")
    yield s3_client


@pytest.fixture(scope="module")
def lx_gcs(settings):
    credentials_file = settings.GOOGLE_APPLICATION_CREDENTIALS
    if not credentials_file:
        warnings.warn("GOOGLE_APPLICATION_CREDENTIALS is not set", UserWarning)
        pytest.skip("GOOGLE_APPLICATION_CREDENTIALS is not set")

    gcs_client = GCSClient()
    if not gcs_client.is_authenticated():
        warnings.warn("GCS client is not authenticated", UserWarning)
        pytest.skip("GCS client is not authenticated")
    yield gcs_client


@pytest.fixture(scope="module")
def test_file_document():
    return "sample_data/documents/hotd.txt"


@pytest.fixture(scope="module")
def test_s3_object(s3, test_file_document, settings):
    bucket_name = settings.S3_TEST_BUCKET
    object_prefix = settings.COLLECTION_DEFAULT_CONFIG["storage_prefix"]
    object_name = os.path.join(object_prefix, "hotd.txt")

    if not bucket_name:
        warnings.warn("S3_TEST_BUCKET is not set", UserWarning)
        pytest.skip("S3_TEST_BUCKET is not set")

    # Upload the test document
    s3.upload_file(test_file_document, bucket_name, object_name)
    print(f"Created test_s3_object: 's3://{bucket_name}/{object_name}'")

    yield bucket_name, object_name

    # Clean up
    s3.delete_object(Bucket=bucket_name, Key=object_name)
    print(f"Deleted test_s3_object: 's3://{bucket_name}/{object_name}'")


@pytest.fixture(scope="module")
def test_gcs_object(gcs, test_file_document, settings):
    bucket_name = settings.GCS_TEST_BUCKET
    object_prefix = settings.COLLECTION_DEFAULT_CONFIG["storage_prefix"]
    object_name = os.path.join(object_prefix, "hotd.txt")

    if not bucket_name:
        warnings.warn("GCS_TEST_BUCKET is not set", UserWarning)
        pytest.skip("GCS_TEST_BUCKET is not set")

    # Upload the test document
    bucket = gcs.bucket(bucket_name)
    blob: storage.blob.Blob = bucket.blob(object_name)
    with open(test_file_document, "rb") as file:
        blob.upload_from_file(file)
    print(f"Created test_gcs_object: 'gs://{bucket_name}/{object_name}'")

    yield bucket_name, object_name

    # Clean up
    blob.delete()
    print(f"Deleted test_gcs_object: 'gs://{bucket_name}/{object_name}'")


# TODO: add async tests
class TestStorageClient:
    """Tests for third-party clients."""

    def test_s3_client(self, s3):
        assert s3 is not None
        buckets = s3.list_buckets()
        assert len(buckets) > 0

    def test_s3v4_client(self, s3v4):
        assert s3v4 is not None
        buckets = s3v4.list_buckets()
        assert len(buckets) > 0

    def test_gcs_client(self, gcs):
        assert gcs is not None
        buckets = list(gcs.list_buckets())
        assert len(buckets) > 0

    def test_generate_signed_url_s3(self, s3, test_s3_object):
        bucket_name, object_name = test_s3_object
        expiration = 3

        s3_signed_url = s3.generate_presigned_url(
            "get_object",
            Params={"Bucket": bucket_name, "Key": object_name},
            ExpiresIn=expiration,
        )
        assert s3_signed_url is not None
        # s3 url can also include the region
        assert s3_signed_url.startswith(f"https://{bucket_name}.s3.")
        assert f".amazonaws.com/{object_name}" in s3_signed_url
        assert signed_url_is_expired(s3_signed_url, svc="Amz") is False
        time.sleep(expiration)
        assert signed_url_is_expired(s3_signed_url, svc="Amz") is True

    def test_generate_signed_url_s3v4(self, s3v4, test_s3_object):
        bucket_name, object_name = test_s3_object
        expiration = 3

        s3v4_signed_url = s3v4.generate_presigned_url(
            "get_object",
            Params={"Bucket": bucket_name, "Key": object_name},
            ExpiresIn=expiration,
        )
        assert s3v4_signed_url is not None
        # s3 url can also include the region
        assert s3v4_signed_url.startswith(f"https://{bucket_name}.s3.")
        assert f".amazonaws.com/{object_name}" in s3v4_signed_url
        assert signed_url_is_expired(s3v4_signed_url, svc="Amz") is False
        time.sleep(expiration)
        assert signed_url_is_expired(s3v4_signed_url, svc="Amz") is True

    def test_generate_signed_url_gcs(self, gcs, test_gcs_object):
        bucket_name, object_name = test_gcs_object
        expiration = 3

        bucket = gcs.bucket(bucket_name)
        blob: storage.blob.Blob = bucket.blob(object_name)
        gcs_signed_url = blob.generate_signed_url(
            expiration=datetime.timedelta(seconds=expiration), version="v4"
        )
        assert gcs_signed_url is not None
        assert gcs_signed_url.startswith(
            f"https://storage.googleapis.com/{bucket_name}/{object_name}"
        )
        assert signed_url_is_expired(gcs_signed_url, svc="Goog") is False
        time.sleep(expiration)
        assert signed_url_is_expired(gcs_signed_url, svc="Goog") is True


# TODO: add async tests
class TestLexyStorageClient:
    """Tests for Lexy storage clients (i.e., `lexy.storage.client.StorageClient`)."""

    def test_lx_s3_client(self, lx_s3):
        assert lx_s3 is not None
        buckets = lx_s3.list_buckets()
        assert len(buckets) > 0

    def test_lx_s3v4_client(self, lx_s3v4):
        assert lx_s3v4 is not None
        buckets = lx_s3v4.list_buckets()
        assert len(buckets) > 0

    def test_lx_gcs_client(self, lx_gcs):
        assert lx_gcs is not None
        buckets = lx_gcs.list_buckets()
        assert len(buckets) > 0

    def test_generate_presigned_url_lx_s3(self, lx_s3, test_s3_object):
        bucket_name, object_name = test_s3_object
        expiration = 3

        s3_signed_url = lx_s3.generate_presigned_url(
            bucket_name, object_name, expiration
        )
        assert s3_signed_url is not None
        # s3 url can also include the region
        assert s3_signed_url.startswith(f"https://{bucket_name}.s3.")
        assert f".amazonaws.com/{object_name}" in s3_signed_url
        assert signed_url_is_expired(s3_signed_url, svc="Amz") is False
        time.sleep(expiration)
        assert signed_url_is_expired(s3_signed_url, svc="Amz") is True

    def test_generate_presigned_url_lx_s3v4(self, lx_s3v4, test_s3_object):
        bucket_name, object_name = test_s3_object
        expiration = 3

        s3v4_signed_url = lx_s3v4.generate_presigned_url(
            bucket_name, object_name, expiration
        )
        assert s3v4_signed_url is not None
        # s3 url can also include the region
        assert s3v4_signed_url.startswith(f"https://{bucket_name}.s3.")
        assert f".amazonaws.com/{object_name}" in s3v4_signed_url
        assert signed_url_is_expired(s3v4_signed_url, svc="Amz") is False
        time.sleep(expiration)
        assert signed_url_is_expired(s3v4_signed_url, svc="Amz") is True

    def test_generate_presigned_url_lx_gcs(self, lx_gcs, test_gcs_object):
        bucket_name, object_name = test_gcs_object
        expiration = 3

        gcs_signed_url = lx_gcs.generate_presigned_url(
            bucket_name, object_name, expiration
        )
        assert gcs_signed_url is not None
        assert gcs_signed_url.startswith(
            f"https://storage.googleapis.com/{bucket_name}/{object_name}"
        )
        assert signed_url_is_expired(gcs_signed_url, svc="Goog") is False
        time.sleep(expiration)
        assert signed_url_is_expired(gcs_signed_url, svc="Goog") is True


class TestConstructStorageKeys:
    """Tests for constructed storage keys."""

    @pytest.mark.asyncio
    async def test_construct_key_for_document_with_document(self):
        document = Document(collection_id="123", document_id="456")
        key = await construct_key_for_document(document=document)
        assert key == "collections/123/documents/456"

    @pytest.mark.asyncio
    async def test_construct_key_for_document_with_collection_and_document_id(self):
        key = await construct_key_for_document(collection_id="123", document_id="456")
        assert key == "collections/123/documents/456"

    @pytest.mark.asyncio
    async def test_construct_key_for_document_with_path_prefix(self):
        document = Document(collection_id="123", document_id="456")
        key = await construct_key_for_document(document=document, path_prefix="public")
        assert key == "public/collections/123/documents/456"

    @pytest.mark.asyncio
    async def test_construct_key_for_document_with_filename(self):
        document = Document(collection_id="123", document_id="456")
        key = await construct_key_for_document(
            document=document, filename="document.jpg"
        )
        assert key == "collections/123/documents/456/document.jpg"

    @pytest.mark.asyncio
    async def test_construct_key_for_document_without_document_or_ids(self):
        with pytest.raises(
            ValueError,
            match="Either a document object or collection_id and document_id must be "
            "provided.",
        ):
            await construct_key_for_document()

    @pytest.mark.asyncio
    async def test_construct_key_for_thumbnail_with_document(self):
        document = Document(collection_id="123", document_id="456")
        key = await construct_key_for_thumbnail((200, 200), document=document)
        assert key == "collections/123/thumbnails/200x200/456"

    @pytest.mark.asyncio
    async def test_construct_key_for_thumbnail_with_collection_and_document_id(self):
        key = await construct_key_for_thumbnail(
            (200, 200), collection_id="123", document_id="456"
        )
        assert key == "collections/123/thumbnails/200x200/456"

    @pytest.mark.asyncio
    async def test_construct_key_for_thumbnail_with_path_prefix(self):
        document = Document(collection_id="123", document_id="456")
        key = await construct_key_for_thumbnail(
            (200, 200), document=document, path_prefix="public"
        )
        assert key == "public/collections/123/thumbnails/200x200/456"

    @pytest.mark.asyncio
    async def test_construct_key_for_thumbnail_with_filename(self):
        document = Document(collection_id="123", document_id="456")
        key = await construct_key_for_thumbnail(
            (200, 200), document=document, filename="thumbnail.jpg"
        )
        assert key == "collections/123/thumbnails/200x200/456/thumbnail.jpg"

    @pytest.mark.asyncio
    async def test_construct_key_for_thumbnail_without_dims(self):
        document = Document(collection_id="123", document_id="456")
        with pytest.raises(ValueError, match="Thumbnail dimensions must be provided."):
            await construct_key_for_thumbnail(None, document=document)

    @pytest.mark.asyncio
    async def test_construct_key_for_thumbnail_without_document_or_ids(self):
        with pytest.raises(
            ValueError,
            match="Either a document object or collection_id and document_id must be "
            "provided.",
        ):
            await construct_key_for_thumbnail((200, 200))



from lexy.models.binding import Binding, BindingCreate, BindingUpdate
from lexy.models.collection import Collection, CollectionCreate, CollectionUpdate
from lexy.models.document import Document, DocumentCreate, DocumentUpdate
from lexy.models.index import Index, IndexCreate, IndexUpdate
from lexy.models.index_record import (
    IndexRecordBase,
    IndexRecordBaseTable,
    IndexRecordCreate,
    IndexRecordUpdate,
)
from lexy.models.transformer import Transformer, TransformerCreate, TransformerUpdate
from lexy.models.user import User, UserCreate, UserUpdate

__all__ = [
    "Binding",
    "BindingCreate",
    "BindingUpdate",
    "Collection",
    "CollectionCreate",
    "CollectionUpdate",
    "Document",
    "DocumentCreate",
    "DocumentUpdate",
    "Index",
    "IndexCreate",
    "IndexUpdate",
    "IndexRecordBase",
    "IndexRecordBaseTable",
    "IndexRecordCreate",
    "IndexRecordUpdate",
    "Transformer",
    "TransformerCreate",
    "TransformerUpdate",
    "User",
    "UserCreate",
    "UserUpdate",
]



from sqlalchemy.engine import Engine
from sqlalchemy.ext.asyncio import AsyncEngine, async_sessionmaker, create_async_engine
from sqlmodel import create_engine
from sqlmodel.ext.asyncio.session import AsyncSession

from lexy.core.config import settings


sync_engine: Engine = create_engine(
    url=settings.sync_database_url, echo=settings.DB_ECHO_LOG, future=True
)

async_engine: AsyncEngine = create_async_engine(
    url=settings.async_database_url, echo=settings.DB_ECHO_LOG, future=True
)

async_session = async_sessionmaker(
    bind=async_engine, class_=AsyncSession, expire_on_commit=False
)


def get_sync_engine() -> Engine:
    return sync_engine


async def get_session() -> AsyncSession:
    async with async_session() as session:
        yield session



import importlib
import logging
import os
import pkgutil
import sys
from functools import lru_cache
from pathlib import Path
from typing import Literal, Optional

from pydantic import (
    DirectoryPath,
    field_validator,
    EmailStr,
    Field,
    SecretStr,
    ValidationInfo,
)
from pydantic_settings import BaseSettings, SettingsConfigDict


# For some reason, this statement initializes logging for the rest of the application.
#  Without it, we don't get log statements for IndexManager.
logging.info("Loading config.py")


def get_transformer_modules(transformer_pkg: str = "lexy.transformers"):
    """Return a list of transformer modules."""
    lexy_transformer_modules = []
    transformer_pkg_m = importlib.import_module(transformer_pkg)
    for m in pkgutil.iter_modules(
        transformer_pkg_m.__path__, prefix=transformer_pkg + "."
    ):
        if not m.ispkg:
            lexy_transformer_modules.append(m.name)
    return lexy_transformer_modules


def expand_transformer_imports(transformer_imports: set[str]) -> set[str]:
    """Expand transformer imports to include all submodules."""
    expanded_imports = set()
    for imp in transformer_imports:
        if "." in imp and imp.rsplit(".", 1)[1] == "*":
            expanded_imports.update(get_transformer_modules(imp.rsplit(".", 1)[0]))
        else:
            expanded_imports.add(imp)
    return expanded_imports


class AppSettings(BaseSettings):
    # FastAPI settings
    TITLE: str = "Lexy Server"
    VERSION: str = "1.0.0"
    SUMMARY: str = "REST API for Lexy Server"
    LICENSE_INFO: dict = {
        "name": "Apache 2.0",
        "url": "https://www.apache.org/licenses/LICENSE-2.0.html",
    }
    OPENAPI_PREFIX: str = ""
    DOCS_URL: str = "/docs"
    REDOC_URL: str = "/redoc"
    OPENAPI_URL: str = "/openapi.json"
    API_PREFIX: str = "/api"
    SERVERS: list[dict] = [{"url": "http://localhost:9900"}]

    # Security settings
    # Uncomment the line below if you want to generate a new secret key every time the
    #  server restarts. Then, to use a fixed key, you would simply set the value of
    #  SECRET_KEY in your .env file. If you uncomment the next line, add
    #  `import secrets` to the top of this file.
    # SECRET_KEY: SecretStr = Field(default_factory=secrets.token_urlsafe, env="SECRET_KEY")
    SECRET_KEY: SecretStr = Field(default="changethis", validation_alias="SECRET_KEY")
    ACCESS_TOKEN_EXPIRE_MINUTES: int = 60 * 24 * 8  # 8 days

    # Database settings
    POSTGRES_USER: str = Field(default="postgres", validation_alias="POSTGRES_USER")
    POSTGRES_PASSWORD: SecretStr = Field(
        default="postgres", validation_alias="POSTGRES_PASSWORD"
    )
    POSTGRES_HOST: str = Field(default="localhost", validation_alias="POSTGRES_HOST")
    POSTGRES_DB: str = Field(default="lexy", validation_alias="POSTGRES_DB")
    DB_ECHO_LOG: bool = False

    # User settings
    FIRST_SUPERUSER_EMAIL: EmailStr = Field(
        "lexy@lexy.ai", validation_alias="FIRST_SUPERUSER_EMAIL"
    )
    FIRST_SUPERUSER_PASSWORD: SecretStr = Field(
        "lexy", validation_alias="FIRST_SUPERUSER_PASSWORD"
    )

    # AWS settings
    AWS_ACCESS_KEY_ID: Optional[str] = Field(
        default=None, validation_alias="AWS_ACCESS_KEY_ID"
    )
    AWS_SECRET_ACCESS_KEY: Optional[SecretStr] = Field(
        default=None, validation_alias="AWS_SECRET_ACCESS_KEY"
    )
    AWS_REGION: Optional[str] = Field(default=None, validation_alias="AWS_REGION")

    # Google Cloud settings
    # Path to a file containing JSON credentials for a service account. Using
    #  Optional[str] because setting to Optional[FilePath] triggers a validation error
    #  if GOOGLE_APPLICATION_CREDENTIALS is an empty string.
    #  If set to devnull, or to an invalid filepath, GOOGLE_APPLICATION_CREDENTIALS
    #  will be set to None.
    GOOGLE_APPLICATION_CREDENTIALS: Optional[str] = Field(
        default=None, validation_alias="GOOGLE_APPLICATION_CREDENTIALS"
    )

    @field_validator("GOOGLE_APPLICATION_CREDENTIALS")
    def check_google_application_credentials(cls, value):
        if not value:
            # Occurs when value is set to an empty string
            return None
        if value == os.path.devnull:
            return None
        if value and not Path(value).is_file():
            logging.warning(
                f"GOOGLE_APPLICATION_CREDENTIALS file '{value}' does not exist. "
                f"Setting value to `None`."
            )
            return None
        if value and Path(value).is_file() and os.path.getsize(value) == 0:
            # For example, when set to 'lexy/docker/null-credentials.json'
            logging.warning(
                f"GOOGLE_APPLICATION_CREDENTIALS file '{value}' is empty. "
                f"Setting value to `None`."
            )
            return None
        return value

    # Storage settings
    DEFAULT_STORAGE_SERVICE: Optional[Literal["s3", "gcs"]] = Field(
        default="s3", validation_alias="DEFAULT_STORAGE_SERVICE"
    )
    DEFAULT_STORAGE_BUCKET: Optional[str] = Field(
        default=None, validation_alias="DEFAULT_STORAGE_BUCKET"
    )
    DEFAULT_STORAGE_PREFIX: Optional[str] = Field(
        default=None, validation_alias="DEFAULT_STORAGE_PREFIX"
    )

    # Default config for Collection objects and images
    COLLECTION_DEFAULT_CONFIG: dict = {"store_files": True, "generate_thumbnails": True}

    @field_validator("COLLECTION_DEFAULT_CONFIG")
    def set_collection_default_config(cls, v: dict, info: ValidationInfo) -> dict:
        if v.get("store_files") is True:
            if not v.get("storage_service"):
                v["storage_service"] = info.data.get("DEFAULT_STORAGE_SERVICE")
            if not v.get("storage_bucket"):
                v["storage_bucket"] = info.data.get("DEFAULT_STORAGE_BUCKET")
            if not v.get("storage_prefix"):
                v["storage_prefix"] = info.data.get("DEFAULT_STORAGE_PREFIX")
        return v

    IMAGE_THUMBNAIL_SIZES: set[tuple] = {
        # (100, 100),
        (200, 200),
    }

    # Transformer settings
    LEXY_SERVER_TRANSFORMER_IMPORTS: set[str] = {
        # 'lexy.transformers.*'
        "lexy.transformers.counter",
        "lexy.transformers.embeddings",
        "lexy.transformers.multimodal",
        "lexy.transformers.openai",
    }
    LEXY_WORKER_TRANSFORMER_IMPORTS: set[str] = {
        # 'lexy.transformers.*'
        "lexy.transformers.counter",
        "lexy.transformers.embeddings",
        "lexy.transformers.multimodal",
        "lexy.transformers.openai",
        "pipelines.*",
    }

    # Pipeline directory (relative to the project root)
    PIPELINE_DIR: DirectoryPath = Field(
        default="pipelines", validation_alias="PIPELINE_DIR"
    )

    @property
    def sync_database_url(self) -> str:
        return (
            f"postgresql://"
            f"{self.POSTGRES_USER}"
            f":{self.POSTGRES_PASSWORD.get_secret_value()}"
            f"@{self.POSTGRES_HOST}"
            f"/{self.POSTGRES_DB}"
        )

    @property
    def async_database_url(self) -> str:
        return (
            f"postgresql+asyncpg://"
            f"{self.POSTGRES_USER}"
            f":{self.POSTGRES_PASSWORD.get_secret_value()}"
            f"@{self.POSTGRES_HOST}"
            f"/{self.POSTGRES_DB}"
        )

    @property
    def app_transformer_imports(self):
        return expand_transformer_imports(self.LEXY_SERVER_TRANSFORMER_IMPORTS)

    @property
    def worker_transformer_imports(self):
        # TODO: Move to a separate property for pipeline imports
        # if self.PIPELINE_DIR:
        #     # Add pipeline imports to the worker transformer imports (i.e., adds
        #     # 'pipelines.*' to the imports)
        #     pipeline_imports = f'{self.PIPELINE_DIR.name}.*'
        #     expanded_set = self.LEXY_WORKER_TRANSFORMER_IMPORTS.union({pipeline_imports})
        #     return expand_transformer_imports(expanded_set)
        return expand_transformer_imports(self.LEXY_WORKER_TRANSFORMER_IMPORTS)

    @field_validator("PIPELINE_DIR", mode="before")
    @classmethod
    def check_pipeline_dir(cls, value):
        if value:
            pipeline_dir = Path(value).resolve()
            if not pipeline_dir.is_dir():
                err_msg = (
                    f"Pipeline directory '{pipeline_dir}' does not exist. Run "
                    f"`lexy init` to initialize your project and create the directory."
                )
                raise ValueError(err_msg)
            # log a warning if pipeline_dir is not in the Python path (Celery can't
            # import pipelines)
            if str(pipeline_dir) not in sys.path and str(pipeline_dir) != os.getcwd():
                logging.warning(
                    f"Pipeline directory '{pipeline_dir}' is not in the Python path. "
                    f"Your pipelines may not be imported correctly."
                )
        return value

    model_config = SettingsConfigDict(
        case_sensitive=True, env_file=".env", env_file_encoding="utf-8", extra="allow"
    )


class DevelopmentAppSettings(AppSettings):
    pass


class ProductionAppSettings(AppSettings):
    @field_validator("SECRET_KEY")
    @classmethod
    def check_secret_key(cls, value):
        if value.get_secret_value() == "changethis":
            logging.error(
                "Using default value of SECRET_KEY in production settings! "
                "Change this value to a secure secret key."
            )
            raise Exception("Using default value of SECRET_KEY in production settings!")
        return value


class TestAppSettings(AppSettings):
    # Database settings
    POSTGRES_DB: str = Field(default="lexy_tests", validation_alias="POSTGRES_TEST_DB")
    DB_ECHO_LOG: bool = False

    # User settings
    # Without `env=` argument, this will revert to the environment value of
    # FIRST_SUPERUSER_EMAIL
    FIRST_SUPERUSER_EMAIL: EmailStr = Field(
        "test@lexy.ai", validation_alias="TEST_SUPERUSER_EMAIL"
    )
    FIRST_SUPERUSER_PASSWORD: SecretStr = Field(
        "test", validation_alias="TEST_SUPERUSER_PASSWORD"
    )

    # Storage settings
    # By default, use the same storage service and bucket for testing, with a different
    # prefix
    DEFAULT_STORAGE_PREFIX: Optional[str] = Field(
        default="lexy_tests", validation_alias="DEFAULT_STORAGE_PREFIX"
    )
    # Bucket used in testing S3 storage, regardless of the default storage service
    S3_TEST_BUCKET: Optional[str] = Field(
        default=None, validation_alias="S3_TEST_BUCKET"
    )
    # Bucket used in testing GCS storage, regardless of the default storage service
    GCS_TEST_BUCKET: Optional[str] = Field(
        default=None, validation_alias="GCS_TEST_BUCKET"
    )

    # Worker settings
    # Remove 'pipelines.*' from worker_transformer_imports - this avoids
    #  ModuleNotFoundError when running tests on lexy-server, which does not have any
    #  of the extra pipeline requirements installed.
    LEXY_WORKER_TRANSFORMER_IMPORTS: set[str] = {
        # 'lexy.transformers.*'
        "lexy.transformers.counter",
        "lexy.transformers.embeddings",
        "lexy.transformers.multimodal",
        "lexy.transformers.openai",
    }


@lru_cache()
def get_settings():
    config_cls_dict = {
        "development": DevelopmentAppSettings,
        "testing": TestAppSettings,
    }
    config_name = os.environ.get("LEXY_CONFIG", "development")
    config_cls = config_cls_dict[config_name]
    return config_cls()


settings = get_settings()



import os
from importlib.util import find_spec

if find_spec("pytest") is None:
    raise ImportError(
        "The 'lexy_tests' package requires additional dependencies. "
        "Please install them using 'pip install lexy[tests]'."
    )

os.environ["LEXY_CONFIG"] = "testing"
os.environ["CELERY_CONFIG"] = "testing"

from lexy_tests.conftest import async_client, client, test_settings  # noqa: E402

__all__ = ["async_client", "client", "test_settings"]



import asyncio
import pytest
from sqlmodel import select

from lexy.models.collection import Collection
from lexy.models.document import Document, DocumentCreate, DocumentUpdate


class TestDocument:
    def test_hello(self):
        assert True

    @pytest.mark.asyncio
    async def test_create_document(self, async_session):
        result = await async_session.exec(
            select(Collection).where(Collection.collection_name == "default")
        )
        default_collection = result.first()
        document = Document(
            content="Test Content", collection_id=default_collection.collection_id
        )

        async_session.add(document)
        await async_session.commit()
        await async_session.refresh(document)
        assert document.content == "Test Content"
        assert document.document_id is not None
        assert document.created_at is not None
        assert document.updated_at is not None

        result = await async_session.exec(select(Document))
        documents = result.all()
        assert len(documents) == 1
        assert documents[0].content == "Test Content"

    @pytest.mark.asyncio
    async def test_get_documents_with_async_client(self, async_client):
        response = await async_client.get(
            "/api/documents",
        )
        assert response.status_code == 200, response.text

        data = response.json()
        assert len(data) == 1
        assert data[0]["content"] == "Test Content"

    @pytest.mark.asyncio
    async def test_add_document(self, async_session, async_client):
        result = await async_session.exec(
            select(Collection).where(Collection.collection_name == "default")
        )
        default_collection = result.first()
        doc = Document(
            content="a shiny new document",
            collection_id=default_collection.collection_id,
        )
        async_session.add(doc)
        await async_session.commit()

        response = await async_client.get(
            f"/api/documents/{doc.document_id}",
        )
        assert response.status_code == 200, response.text

        data = response.json()
        assert data["content"] == doc.content
        assert data["document_id"] == str(doc.document_id)
        assert data["created_at"] == doc.created_at.isoformat().replace("+00:00", "Z")
        assert data["updated_at"] == doc.updated_at.isoformat().replace("+00:00", "Z")

    @pytest.mark.asyncio
    async def test_add_documents(self, async_session, async_client):
        result = await async_session.exec(
            select(Collection).where(Collection.collection_name == "code")
        )
        code_collection = result.first()
        doc1 = Document(
            content="import this", collection_id=code_collection.collection_id
        )
        doc2 = Document(
            content="export that", collection_id=code_collection.collection_id
        )
        async_session.add(doc1)
        async_session.add(doc2)
        await async_session.commit()

        response = await async_client.get(
            "/api/documents", params={"collection_name": "code"}
        )
        assert response.status_code == 200, response.text

        data = response.json()
        assert len(data) == 2

        assert data[0]["content"] == "import this"
        assert data[0]["collection_id"] == code_collection.collection_id
        assert data[0]["document_id"] == str(doc1.document_id)
        assert data[0]["created_at"] == doc1.created_at.isoformat().replace(
            "+00:00", "Z"
        )
        assert data[0]["updated_at"] == doc1.updated_at.isoformat().replace(
            "+00:00", "Z"
        )

        assert data[1]["content"] == "export that"
        assert data[1]["collection_id"] == code_collection.collection_id
        assert data[1]["document_id"] == str(doc2.document_id)
        assert data[1]["created_at"] == doc2.created_at.isoformat().replace(
            "+00:00", "Z"
        )
        assert data[1]["updated_at"] == doc2.updated_at.isoformat().replace(
            "+00:00", "Z"
        )

    @pytest.mark.asyncio
    async def test_document_crud(self, async_session):
        # get default collection
        result = await async_session.exec(
            select(Collection).where(Collection.collection_name == "default")
        )
        default_collection = result.first()

        # create document
        document = DocumentCreate(content="Test Document CRUD")
        db_document = Document(
            **document.model_dump(), collection_id=default_collection.collection_id
        )
        async_session.add(db_document)
        await async_session.commit()
        await async_session.refresh(db_document)
        assert db_document.content == "Test Document CRUD"
        assert db_document.collection_id == default_collection.collection_id
        assert db_document.document_id is not None
        assert db_document.created_at is not None
        assert db_document.updated_at is not None
        doc_id = db_document.document_id

        # get document
        result = await async_session.exec(
            select(Document).where(Document.document_id == doc_id)
        )
        documents = result.all()
        assert len(documents) == 1
        assert documents[0].content == "Test Document CRUD"

        # update document
        document_update = DocumentUpdate(content="Test Document CRUD Updated")
        document_data = document_update.model_dump(exclude_unset=True)
        for key, value in document_data.items():
            setattr(db_document, key, value)
        async_session.add(db_document)
        await async_session.commit()
        await async_session.refresh(db_document)
        assert db_document.content == "Test Document CRUD Updated"
        assert db_document.created_at is not None
        assert db_document.updated_at > db_document.created_at
        result = await async_session.exec(
            select(Document).where(Document.document_id == doc_id)
        )
        documents = result.all()
        assert len(documents) == 1
        assert documents[0].content == "Test Document CRUD Updated"
        assert documents[0].created_at is not None
        assert documents[0].updated_at > documents[0].created_at

        # delete document
        result = await async_session.exec(
            select(Document).where(Document.document_id == doc_id)
        )
        doc_to_delete = result.first()
        assert doc_to_delete is not None
        await async_session.delete(doc_to_delete)
        await async_session.commit()

        result = await async_session.exec(
            select(Document).where(Document.document_id == doc_id)
        )
        doc_to_delete = result.first()
        assert doc_to_delete is None

    @pytest.mark.asyncio
    async def test_add_documents_with_async_client(
        self, async_client, celery_app, celery_worker
    ):
        response = await async_client.post(
            "/api/documents",
            json=[{"content": "hello there!"}],
        )
        assert response.status_code == 201, response.text

        data = response.json()
        assert len(data) == 1
        assert set(data[0].keys()) == {"document", "tasks"}
        assert data[0]["document"]["content"] == "hello there!"
        doc_id = data[0]["document"]["document_id"]

        # wait for the celery worker to finish the task
        await asyncio.sleep(1)

        # retrieve the index records for default_text_embeddings
        records_response = await async_client.get(
            "/api/indexes/default_text_embeddings/records",
        )
        assert records_response.status_code == 200, records_response.text
        records_data = records_response.json()
        assert len(records_data) == 1
        assert records_data[0]["document_id"] == doc_id
        index_record_id = records_data[0]["index_record_id"]

        # retrieve the index record by id
        index_record_response = await async_client.get(
            f"/api/indexes/default_text_embeddings/records/{index_record_id}",
        )
        assert index_record_response.status_code == 200, index_record_response.text
        index_record_data = index_record_response.json()
        assert index_record_data["document_id"] == doc_id
        assert index_record_data["index_record_id"] == index_record_id



from datetime import datetime, timedelta, timezone
from typing import Literal
from urllib.parse import parse_qs, urlparse


def presigned_url_is_expired(url: str, storage_service: str = "s3") -> bool:
    if storage_service == "s3":
        return signed_url_is_expired(url, svc="Amz")
    elif storage_service == "gcs":
        return signed_url_is_expired(url, svc="Goog")
    elif storage_service == "azure":
        raise NotImplementedError
    else:
        raise ValueError(
            f"Unsupported storage service: {storage_service} - "
            f"must be one of the following: s3, gcs, azure"
        )


def signed_url_is_expired(url: str, svc: Literal["Amz", "Goog"]):
    """Checks if a signed URL (V2 or V4) has expired.

    Args:
        url (str): The signed URL string.
        svc (Literal['Amz', 'Goog']): The service query parameter used to check for
            expiration in V4 signatures. Either 'Amz' for AWS or 'Goog' for GCP.

    Returns:
        True if the signed URL is expired, False otherwise.

    Raises:
        ValueError: If no expiration found in signed url.
    """
    query_params = parse_qs(urlparse(url).query)
    # V4 signature
    if f"X-{svc}-Expires" in query_params:
        expires_in = int(query_params[f"X-{svc}-Expires"][0])
        issued_at_dt = datetime.fromisoformat(query_params[f"X-{svc}-Date"][0])
        expires_at_dt = issued_at_dt + timedelta(seconds=expires_in)
        return expires_at_dt < datetime.now(tz=timezone.utc)
    # V2 signature
    elif "Expires" in query_params:
        expires_at = int(query_params["Expires"][0])
        expires_at_dt = datetime.fromtimestamp(expires_at, tz=timezone.utc)
        return expires_at_dt < datetime.now(tz=timezone.utc)
    else:
        raise ValueError("No expiration found in signed url")



from lexy.models.document import DocumentBase
from lexy.transformers import lexy_transformer


@lexy_transformer(name="text.counter.word_counter")
def word_counter(text: str) -> tuple[int, str]:
    """Testing a transformer."""
    if isinstance(text, DocumentBase):
        text = text.content
    words = text.split()
    word_count = len(words)
    longest_word = max(words, key=len)
    return word_count, longest_word



from io import BytesIO

from fastapi import APIRouter, Depends, HTTPException, Query, status, UploadFile
from PIL import Image
from sqlmodel import delete, select
from sqlmodel.ext.asyncio.session import AsyncSession

from lexy import crud
from lexy.db.session import get_session
from lexy.storage.client import (
    construct_key_for_document,
    construct_key_for_thumbnail,
    generate_signed_urls_for_document,
    get_storage_client,
    StorageClient,
)
from lexy.models.document import Document, DocumentCreate, DocumentUpdate
from lexy.core.config import settings
from lexy.core.events import generate_tasks_for_document


router = APIRouter()


@router.get(
    "/documents",
    response_model=list[Document],
    status_code=status.HTTP_200_OK,
    name="get_documents",
    description="Get documents in a collection",
)
async def get_documents(
    collection_name: str = "default",
    limit: int = Query(100, gt=0, le=1000),
    offset: int = 0,
    session: AsyncSession = Depends(get_session),
) -> list[Document]:
    # get the collection
    collection = await crud.get_collection_by_name(
        session=session, collection_name=collection_name
    )
    if not collection:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Collection not found"
        )

    # get documents
    result = await session.exec(
        select(Document)
        .where(Document.collection_id == collection.collection_id)
        .limit(limit)
        .offset(offset)
    )
    documents = result.all()
    return documents


@router.post(
    "/documents",
    status_code=status.HTTP_201_CREATED,
    name="add_documents",
    description="Add documents to a collection",
)
async def add_documents(
    documents: list[DocumentCreate],
    collection_name: str = "default",
    session: AsyncSession = Depends(get_session),
) -> list[dict]:
    # get the collection
    collection = await crud.get_collection_by_name(
        session=session, collection_name=collection_name
    )
    if not collection:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Collection not found"
        )

    docs_added = []
    for doc in documents:
        # Check if a document_id is provided and if it already exists
        if doc.document_id:
            existing_doc = await session.get(Document, doc.document_id)
            if existing_doc:
                raise HTTPException(
                    status_code=400,
                    detail={
                        "msg": f"A document with this ID already exists: "
                        f"{doc.document_id}.",
                        "document_id": str(doc.document_id),
                    },
                )
        document = Document(**doc.model_dump(), collection_id=collection.collection_id)
        session.add(document)
        await session.commit()
        await session.refresh(document)
        tasks = await generate_tasks_for_document(document)
        docs_added.append({"document": document, "tasks": tasks})
    return docs_added


# TODO: refactor logic to combine with `lexy.api.endpoints.collections.upload_collection_documents`
@router.post(
    "/documents/upload",
    status_code=status.HTTP_201_CREATED,
    name="upload_documents",
    description="Upload documents to a collection",
)
async def upload_documents(
    files: list[UploadFile],
    collection_name: str = "default",
    session: AsyncSession = Depends(get_session),
    storage_client: StorageClient = Depends(get_storage_client),
) -> list[dict]:
    # get the collection
    collection = await crud.get_collection_by_name(
        session=session, collection_name=collection_name
    )
    if not collection:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Collection not found"
        )
    collection_id = collection.collection_id
    # TODO: get storage client from collection config
    # storage_client = get_storage_client(service=collection.config.get('storage_service', None))
    if collection.config.get("store_files") and not storage_client:
        storage_service = collection.config.get("storage_service", None)
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"Storage client not configured (service: {storage_service})",
        )
    storage_bucket = collection.config.get(
        "storage_bucket", settings.DEFAULT_STORAGE_BUCKET
    )
    if collection.config.get("store_files") and not storage_bucket:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Storage bucket not configured for this collection",
        )
    storage_prefix = collection.config.get("storage_prefix", None)

    docs_uploaded = []

    for file in files:
        file_dict = {
            "content_type": file.content_type,
            "filename": file.filename,
            "size": file.size,
        }

        # TODO: replace file.filename with the unique document id - will require saving
        #  the document in the DB first but for now, we're using the filename as the
        #  document_id, which is equivalent to the following:
        #    document_key = f"collections/{collection_id}/documents/{file.filename}"
        # uncomment the following line when ready
        # document_key = await construct_key_for_document(
        #     document=document,
        #     path_prefix=storage_prefix,
        #     filename=file.filename
        # )
        document_key = await construct_key_for_document(
            collection_id=collection_id,
            document_id=file.filename,
            path_prefix=storage_prefix,
        )

        # TODO: move this to a separate parsing function - don't read into memory if
        #  not parsing
        file_content = await file.read()
        file_in_memory = BytesIO(file_content)
        file_in_memory.seek(0)

        # "pre-processing" based on file type
        if file.content_type.startswith("image/"):
            # logic for image files
            file_dict["type"] = "image"
            doc_content = f"<Image({file.filename})>"
        elif file.content_type.startswith("text/"):
            # logic for text files
            file_dict["type"] = "text"
            doc_content = file_content.decode("utf-8")
        elif file.content_type.startswith("application/pdf"):
            # logic for pdf files
            file_dict["type"] = "pdf"
            doc_content = f"<PDF({file.filename})>"
        elif (
            file.content_type
            == "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
        ):
            # logic for docx files
            file_dict["type"] = "docx"
            doc_content = f"<DOCX({file.filename})>"
        elif file.content_type.startswith("video/"):
            # logic for video files
            file_dict["type"] = "video"
            doc_content = f"<Video({file.filename})>"
        else:
            # logic for all other files
            file_dict["type"] = "file"
            doc_content = f"<File({file.filename})>"

        # check for an existing doc based on content
        docs = await crud.get_documents_by_collection_id_and_content(
            session=session, collection_id=collection_id, content=doc_content
        )
        if docs:
            file_dict["document"] = docs[0]
            docs_uploaded.append(file_dict)
            continue

        # "post-processing" based on file type
        if file.content_type.startswith("image/"):
            # all we're doing here is getting the width and height - if we're
            # generating thumbnails, we'll use this instance as the input
            img = Image.open(file_in_memory)
            width, height = img.size
            file_dict["image"] = {"width": width, "height": height}

            # generate thumbnails
            if collection.config.get("generate_thumbnails") and collection.config.get(
                "store_files"
            ):
                # generate thumbnails and upload to storage
                thumbnails = {}
                for dims in settings.IMAGE_THUMBNAIL_SIZES:
                    thumbnail_key = await construct_key_for_thumbnail(
                        dims=dims,
                        collection_id=collection_id,
                        document_id=file.filename,
                        path_prefix=storage_prefix,
                    )
                    img_copy = img.copy()
                    img_copy.thumbnail(dims, Image.Resampling.LANCZOS)
                    img_byte_arr = BytesIO()
                    img_copy.save(img_byte_arr, format=img.format or "JPEG")
                    # Note: img_byte_arr.seek(0) is run in storage_client.upload_object
                    # by default `rewind=True`
                    storage_thumbnail_meta = storage_client.upload_object(
                        img_byte_arr, storage_bucket, thumbnail_key
                    )
                    thumbnails[f"{dims[0]}x{dims[1]}"] = storage_thumbnail_meta
                file_dict["image"]["thumbnails"] = thumbnails
        elif (
            file.content_type
            == "application/vnd.openxmlformats-officedocument.wordprocessingml.document"
        ):
            # # example processing for docx
            # import docx
            # doc = docx.Document(BytesIO(file.file.read()))
            # num_paragraphs = len(doc.paragraphs)
            # num_tables = len(doc.tables)
            pass

        # Upload the file to storage
        if collection.config.get("store_files"):
            storage_document_meta = storage_client.upload_object(
                file_in_memory, storage_bucket, document_key
            )
            file_dict.update(storage_document_meta)

        document = Document(
            content=doc_content, meta=file_dict, collection_id=collection_id
        )
        session.add(document)
        await session.commit()
        await session.refresh(document)

        # generate tasks
        tasks = await generate_tasks_for_document(
            document, storage_client=storage_client
        )
        file_dict["document"] = document
        file_dict["tasks"] = tasks

        docs_uploaded.append(file_dict)

    return docs_uploaded


@router.delete(
    "/documents",
    status_code=status.HTTP_200_OK,
    name="bulk_delete_documents",
    description="Bulk delete all documents in a collection",
)
async def bulk_delete_documents(
    collection_name: str, session: AsyncSession = Depends(get_session)
) -> dict:
    # get the collection
    collection = await crud.get_collection_by_name(
        session=session, collection_name=collection_name
    )
    if not collection:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Collection not found"
        )
    # delete documents
    statement = delete(Document).where(
        Document.collection_id == collection.collection_id
    )
    result = await session.exec(statement)
    deleted_count = result.rowcount
    await session.commit()
    return {
        "msg": "Documents deleted",
        "collection_id": collection.collection_id,
        "deleted_count": deleted_count,
    }


@router.get(
    "/documents/{document_id}",
    response_model=Document,
    status_code=status.HTTP_200_OK,
    name="get_document",
    description="Get a document",
)
async def get_document(
    document_id: str, session: AsyncSession = Depends(get_session)
) -> Document:
    result = await session.exec(
        select(Document).where(Document.document_id == document_id)
    )
    document = result.first()
    if not document:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Document not found"
        )
    return document


@router.get(
    "/documents/{document_id}/urls",
    status_code=status.HTTP_200_OK,
    name="get_document_urls",
    description="Get presigned URLs for a document",
)
async def get_document_urls(
    document_id: str,
    expiration: int = Query(3600, gt=0, le=3600),
    session: AsyncSession = Depends(get_session),
    storage_client: StorageClient = Depends(get_storage_client),
) -> dict:
    result = await session.exec(
        select(Document).where(Document.document_id == document_id)
    )
    document = result.first()
    if not document:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Document not found"
        )
    # TODO: get storage client from document config
    if document.is_stored_object and not storage_client:
        storage_service = document.meta.get("storage_service")
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail=f"Storage client not configured (service: {storage_service})",
        )
    return generate_signed_urls_for_document(document, storage_client, expiration)


@router.patch(
    "/documents/{document_id}",
    status_code=status.HTTP_200_OK,
    name="update_document",
    description="Update a document",
)
async def update_document(
    document_id: str,
    document: DocumentUpdate,
    session: AsyncSession = Depends(get_session),
    storage_client: StorageClient = Depends(get_storage_client),
) -> dict:
    result = await session.exec(
        select(Document).where(Document.document_id == document_id)
    )
    db_document = result.first()
    if not db_document:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Document not found"
        )
    document_data = document.model_dump(exclude_unset=True)
    for key, value in document_data.items():
        setattr(db_document, key, value)
    session.add(db_document)
    await session.commit()
    await session.refresh(db_document)
    tasks = await generate_tasks_for_document(
        db_document, storage_client=storage_client
    )
    return {"document": db_document, "tasks": tasks}


@router.delete(
    "/documents/{document_id}",
    status_code=status.HTTP_200_OK,
    name="delete_document",
    description="Delete a document",
)
async def delete_document(
    document_id: str, session: AsyncSession = Depends(get_session)
) -> dict:
    result = await session.exec(
        select(Document).where(Document.document_id == document_id)
    )
    document = result.first()
    if not document:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Document not found"
        )
    await session.delete(document)
    await session.commit()
    return {"msg": "Document deleted", "document_id": document_id}



import contextlib
import importlib
import logging

from fastapi import FastAPI
from fastapi.middleware.cors import CORSMiddleware

from lexy.core.celery_app import create_celery
from lexy.core.config import settings
from lexy.api.deps import index_manager
from lexy.api.router import lexy_api

# Import transformer modules so that celery client can find them
for module in settings.app_transformer_imports:
    importlib.import_module(module)

# Create logger
logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


@contextlib.asynccontextmanager
async def app_lifespan(app: FastAPI):
    # NOTE: anything run here isn't run by celery workers
    logger.info("Starting FastAPI app")
    index_manager.create_index_models()
    logger.debug("(main.app_lifespan) Created index models")
    logger.debug(f"(main.app_lifespan) {index_manager = }")
    logger.debug(f"(main.app_lifespan) {index_manager.index_models = }")
    # Uncomment the next line to access the index_manager from the app state
    # app.state.index_manager = index_manager
    yield
    logger.info("Stopping FastAPI app")


# Create FastAPI app
app = FastAPI(
    title=settings.TITLE,
    version=settings.VERSION,
    summary=settings.SUMMARY,
    license_info=settings.LICENSE_INFO,
    openapi_prefix=settings.OPENAPI_PREFIX,
    docs_url=settings.DOCS_URL,
    openapi_url=settings.OPENAPI_URL,
    servers=settings.SERVERS,
    lifespan=app_lifespan,
)
app.include_router(lexy_api, prefix=settings.API_PREFIX)

# Create Celery app
app.celery_app = create_celery()
celery = app.celery_app

# Add CORS middleware
origins = [
    "http://localhost",
    "http://localhost:3000",
]
app.add_middleware(
    CORSMiddleware,
    allow_origins=origins,
    allow_credentials=True,
    allow_methods=["*"],  # allow all methods
    allow_headers=["*"],  # allow all headers
)



import pytest


class TestFixtures:
    """Test the fixtures in `lexy_tests/conftest.py`."""

    def test_hello(self):
        assert True

    def test_get_root_with_client(self, client):
        response = client.get("/api")
        assert response.status_code == 200
        assert response.json() == {"Say": "Hello!"}

    @pytest.mark.asyncio
    async def test_get_root_with_client(self, client):  # noqa: F811
        response = client.get("/api")
        assert response.status_code == 200
        assert response.json() == {"Say": "Hello!"}

    @pytest.mark.asyncio
    async def test_get_root_with_async_client(self, async_client):
        response = await async_client.get(
            # Trailing slash is required on the root path only.
            # See https://stackoverflow.com/a/70354027
            "/api/",
        )
        assert response.status_code == 200
        assert response.json() == {"Say": "Hello!"}

    def test_get_ping_with_client(self, client):
        response = client.get(
            "/api/ping",
        )
        assert response.status_code == 200, response.text
        data = response.json()
        assert data == {"ping": "pong!"}

    @pytest.mark.asyncio
    async def test_get_ping_with_client(self, client):  # noqa: F811
        response = client.get(
            "/api/ping",
        )
        assert response.status_code == 200, response.text
        data = response.json()
        assert data == {"ping": "pong!"}

    @pytest.mark.asyncio
    async def test_get_ping_with_async_client(self, async_client):
        response = await async_client.get(
            "/api/ping",
        )
        assert response.status_code == 200, response.text
        data = response.json()
        assert data == {"ping": "pong!"}

    def test_get_db_endpoint_with_client(self, client):
        response = client.get(
            "/api/collections",
        )
        assert response.status_code == 200, response.text
        data = response.json()
        assert len(data) > 1

    @pytest.mark.asyncio
    async def test_get_db_endpoint_with_client(self, client):  # noqa: F811
        response = client.get(
            "/api/collections",
        )
        assert response.status_code == 200, response.text
        data = response.json()
        assert len(data) > 1

    @pytest.mark.asyncio
    async def test_get_db_endpoint_with_async_client(self, async_client):
        response = await async_client.get(
            "/api/collections",
        )
        assert response.status_code == 200, response.text
        data = response.json()
        assert len(data) > 1



from datetime import datetime
from typing import Optional

from pydantic import EmailStr
from sqlalchemy import Column, DateTime, func
from sqlmodel import AutoString, Field, SQLModel

from lexy.core.security import get_password_hash


class UserBase(SQLModel):
    email: EmailStr = Field(index=True, nullable=False, unique=True, sa_type=AutoString)
    full_name: Optional[str] = None


class User(UserBase, table=True):
    __tablename__ = "users"
    user_id: int = Field(default=None, primary_key=True, index=True)
    hashed_password: str
    is_active: bool = Field(default=True)
    is_superuser: bool = Field(default=False)
    created_at: datetime = Field(
        sa_column=Column(
            DateTime(timezone=True), nullable=False, server_default=func.now()
        ),
    )
    updated_at: datetime = Field(
        sa_column=Column(
            DateTime(timezone=True),
            nullable=False,
            server_default=func.now(),
            onupdate=func.now(),
        ),
    )

    @classmethod
    def create(cls, email: str, password: str, **kwargs):
        hashed_password = get_password_hash(password)
        return cls(email=email, hashed_password=hashed_password, **kwargs)


# Properties to receive via API on creation
class UserCreate(UserBase):
    email: EmailStr
    password: str


# Additional properties to return via API
class UserRead(UserBase):
    user_id: int
    is_active: bool
    is_superuser: bool
    created_at: datetime
    updated_at: datetime


# Properties to receive via API on update
class UserUpdate(UserBase):
    full_name: Optional[str] = None
    is_active: Optional[bool] = None
    is_superuser: Optional[bool] = None



import pytest

from pydantic_core import ValidationError

from lexy.models import Document
from lexy.schemas.filters import Filter, FilterCondition, Operation, filter_documents


documents = [
    Document(content="", meta={"size": 10000, "type": "image", "category": "photo"}),
    Document(content="this is my text content"),
    Document(
        content="",
        meta={
            "size": 50000,
            "type": "video",
            "url": "https://www.youtube.com/watch?v=dQw4w9WgXcQ",
        },
    ),
    Document(
        content="",
        meta={"size": 12345, "type": "pdf", "url": "https://www.reddit.com/reddit.pdf"},
    ),
]


class TestFilters:
    def test_hello(self):
        assert True

    def test_filter_documents(self):
        filter_obj = Filter(
            conditions=[
                FilterCondition(
                    field="content", operation=Operation.CONTAINS, value="text"
                )
            ]
        )
        filtered_docs = list(filter_documents(documents, filter_obj))
        assert len(filtered_docs) == 1
        assert filtered_docs[0].content == "this is my text content"

    def test_filter_documents_with_multiple_conditions(self):
        filter_obj = Filter(
            conditions=[
                FilterCondition(
                    field="content", operation=Operation.CONTAINS, value="text"
                ),
                FilterCondition(
                    field="meta.size", operation=Operation.GREATER_THAN, value=10000
                ),
            ]
        )
        filtered_docs = list(filter_documents(documents, filter_obj))
        assert len(filtered_docs) == 0

    def test_filter_documents_with_in_condition(self):
        filter_obj = Filter.model_validate(
            {
                "conditions": [
                    {"field": "meta.size", "operation": "less_than", "value": 30000},
                    {
                        "field": "meta.type",
                        "operation": "in",
                        "value": ["image", "video"],
                    },
                ],
                "combination": "AND",
            }
        )
        filtered_docs = list(filter_documents(documents, filter_obj))
        assert len(filtered_docs) == 1
        assert filtered_docs[0] == documents[0]

    def test_filter_documents_with_exclude_in_condition(self):
        filter_obj = Filter.model_validate(
            {
                "conditions": [
                    {"field": "meta.size", "operation": "less_than", "value": 30000},
                    {
                        "field": "meta.type",
                        "operation": "in",
                        "value": ["image", "video"],
                        "negate": True,
                    },
                ],
                "combination": "AND",
            }
        )
        filtered_docs = list(filter_documents(documents, filter_obj))
        assert len(filtered_docs) == 1
        assert filtered_docs[0] == documents[3]

    def test_filter_documents_with_isnull_condition(self):
        filter_obj = Filter.model_validate(
            {
                "conditions": [
                    {"field": "meta.size", "operation": "equals", "value": None}
                ],
                "combination": "AND",
            }
        )
        filtered_docs = list(filter_documents(documents, filter_obj))
        assert len(filtered_docs) == 1
        assert filtered_docs[0] == documents[1]

    def test_filter_documents_with_notnull_condition(self):
        filter_obj = Filter.model_validate(
            {
                "conditions": [
                    {
                        "field": "meta.size",
                        "operation": "equals",
                        "value": None,
                        "negate": True,
                    }
                ],
                "combination": "AND",
            }
        )
        filtered_docs = list(filter_documents(documents, filter_obj))
        assert len(filtered_docs) == 3
        assert filtered_docs[0] == documents[0]
        assert filtered_docs[1] == documents[2]
        assert filtered_docs[2] == documents[3]

    def test_filter_with_invalid_value_type(self):
        with pytest.raises(ValidationError) as exc_info:
            Filter.model_validate(
                {
                    "conditions": [
                        {
                            "field": "meta.size",
                            "operation": "less_than",
                            "value": "hello",
                        }
                    ],
                    "combination": "AND",
                }
            )
        assert exc_info.type == ValidationError
        val_err = exc_info.value
        assert isinstance(val_err, ValidationError)
        errors = val_err.errors()
        assert len(errors) == 1
        assert errors[0]["type"] == "value_error"
        assert errors[0]["loc"] == ("conditions", 0, "value")
        assert (
            errors[0]["msg"]
            == "Value error, Value must be a number for operation 'less_than'"
        )
        assert errors[0]["input"] == "hello"



"""
Celery tasks for lexy storage.

Sources:
    - Task instantiation:
        - https://stackoverflow.com/a/17224993
        - https://docs.celeryq.dev/en/latest/userguide/tasks.html#instantiation
    - Extending the storage backend:
        - https://stackoverflow.com/questions/60281347/is-it-possible-to-extend-celery-so-results-would-be-store-to-several-mongodb-co
        - https://docs.celeryq.dev/en/stable/userguide/configuration.html#override-backends
"""

from typing import Any
from uuid import UUID

import numpy as np
import torch
from celery import current_app as celery, Task
from celery.utils.log import get_logger, get_task_logger
from sqlalchemy.orm import sessionmaker

from lexy.api.deps import index_manager
from lexy.db.session import sync_engine


logger = get_logger(__name__)
task_logger = get_task_logger(__name__)
SyncSessionLocal = sessionmaker(autocommit=False, autoflush=False, bind=sync_engine)


class DatabaseTask(Task):
    _db = None
    _index_manager = index_manager

    @property
    def db(self):
        if self._db is None:
            self._db = SyncSessionLocal()
        return self._db

    @property
    def index_manager(self):
        if self._index_manager is None:
            task_logger.debug(
                "_index_manager is None - reloading and assigning index manager"
            )
            self._index_manager = index_manager
            task_logger.debug(
                f"Assigned index manager with index models: "
                f"{self._index_manager.index_models}"
            )
        if self._index_manager.index_models == {}:
            task_logger.debug(
                "index_manager.index_models is empty - running "
                "index_manager.create_index_models()"
            )
            self._index_manager.create_index_models()
            task_logger.debug(
                f"index_manager.index_models: {self._index_manager.index_models}"
            )
        return self._index_manager

    def restart_worker(self, msg):
        """Restart the celery worker of the current task."""
        task_logger.info(f"Called restart_worker with message: {msg}")
        task_logger.debug("---------------- prior to restart ----------------")
        self.print_index_manager_models()
        task_logger.debug(
            f"Broadcasting signal 'pool_restart' to celery worker "
            f"{self.request.hostname}"
        )
        response = self.app.control.broadcast(
            "pool_restart",
            arguments={
                "reload": True,
                "modules": ["lexy.api.deps", "lexy.core.celery_tasks"],
            },
            destination=[self.request.hostname],
            reply=True,
            timeout=3,
        )
        self.reset_index_manager()
        task_logger.debug("---------------- after restart ----------------")
        self.print_index_manager_models()
        return response

    def reset_index_manager(self):
        task_logger.debug("resetting index manager")
        self._index_manager = None

    def print_index_manager_models(self):
        sorted_models = sorted(
            self.index_manager.index_models.items(), key=lambda x: x[0]
        )
        print(
            "index_manager.index_models:\n",
            "\n".join([f"{k}: {v}" for k, v in sorted_models]),
        )


def convert_arrays_to_lists(obj):
    """Convert numpy arrays and PyTorch tensors to lists in a dictionary or list of
    dictionaries.

    Args:
        obj: dictionary or list of dictionaries to convert

    Returns:
        Same type as obj: dictionary or list of dictionaries with numpy arrays and
        PyTorch tensors converted to lists

    Examples:
        >>> convert_arrays_to_lists({
        ...     "a": np.array([1, 2, 3]),
        ...     "b": {"c": torch.tensor([4, 5, 6])}
        ... })
        {'a': [1, 2, 3], 'b': {'c': [4, 5, 6]}}

        >>> convert_arrays_to_lists([
        ...     {"a": torch.tensor([1, 2, 3])},
        ...     {"b": np.array([4, 5, 6])}
        ... ])
        [{'a': [1, 2, 3]}, {'b': [4, 5, 6]}]
    """
    if isinstance(obj, dict):
        for key, value in obj.items():
            if isinstance(value, (np.ndarray, torch.Tensor)):
                obj[key] = value.tolist()
            elif isinstance(value, (dict, list)):
                obj[key] = convert_arrays_to_lists(value)
    elif isinstance(obj, list):
        return [convert_arrays_to_lists(item) for item in obj]
    return obj


@celery.task(base=DatabaseTask, bind=True, name="lexy.db.save_result_to_index")
def save_result_to_index(self, res, document_id, text, index_id):
    """Save the result of a transformer to an index."""
    task_logger.debug(
        f"Starting DB task 'save_result_to_index' for index {index_id} "
        f"with task ID {self.request.id} and parent task ID {self.request.parent_id}"
    )

    try:
        # noinspection PyPep8Naming
        IndexClass = self.index_manager.index_models[index_id]
    except KeyError:
        print(
            f"index model for {index_id} doesn't exist - will attempt to refresh index "
            f"manager"
        )
        self.reset_index_manager()
        # noinspection PyPep8Naming
        IndexClass = self.index_manager.index_models[index_id]

    self.db.add(
        # TODO: IndexClass needs to use kwargs from `index_fields`
        IndexClass(
            document_id=document_id,
            embedding=res.tolist(),
            text=text,
            task_id=self.request.parent_id,
        )
    )
    try:
        self.db.commit()
    except Exception as e:
        self.db.rollback()
        raise e


@celery.task(base=DatabaseTask, bind=True, name="lexy.db.save_records_to_index")
def save_records_to_index(
    self,
    records: list[dict[str, Any]],
    document_id: UUID,
    text: str,
    index_id: str,
    binding_id: int = None,
):
    """Save the output of a transformer to an index."""
    task_logger.debug(
        f"Starting DB task 'save_records_to_index' for index {index_id} "
        f"with task ID {self.request.id} and parent task ID {self.request.parent_id}"
    )

    try:
        # noinspection PyPep8Naming
        IndexClass = self.index_manager.index_models[index_id]
    except KeyError:
        task_logger.warning(
            f"index model for {index_id} doesn't exist - will attempt to refresh index "
            f"manager"
        )
        self.reset_index_manager()
        # noinspection PyPep8Naming
        IndexClass = self.index_manager.index_models[index_id]

    for record in records:
        record = convert_arrays_to_lists(record)
        self.db.add(
            IndexClass(
                document_id=document_id,
                text=text,
                task_id=self.request.parent_id,
                binding_id=binding_id,
                **record,
            )
        )
    try:
        self.db.commit()
    except Exception as e:
        self.db.rollback()
        raise e


@celery.task(base=DatabaseTask, bind=True, name="lexy.db.restart_db_worker")
def restart_db_worker(self, msg):
    """Restart the celery worker of the current task."""
    response = self.restart_worker(msg=msg)
    task_logger.info(f"restart_db_worker response: {response}")
    return response



from lexy.models.document import DocumentBase
from lexy.transformers import lexy_transformer

import torch
from PIL.Image import Image
from transformers import CLIPProcessor, CLIPModel

torch.set_num_threads(1)
model = CLIPModel.from_pretrained("openai/clip-vit-base-patch32")
processor = CLIPProcessor.from_pretrained("openai/clip-vit-base-patch32")

# move model to device if possible
device = "cuda" if torch.cuda.is_available() else "cpu"
model.to(device)


@lexy_transformer(name="image.embeddings.clip")
def image_embeddings_clip(
    images: Image | DocumentBase | list[Image | DocumentBase],
) -> torch.Tensor:
    """Embed images using CLIP.

    Args:
        images: A single image or DocumentBase instance, or a list of images or
            DocumentBase instances to embed.

    Returns:
        torch.Tensor: The embeddings of the provided images.
    """

    if isinstance(images, DocumentBase):
        images = images.image
    elif isinstance(images, list):
        images = [i.image if isinstance(i, DocumentBase) else i for i in images]

    image_batch = processor(text=None, images=images, return_tensors="pt")[
        "pixel_values"
    ].to(device)

    if isinstance(images, list):
        return model.get_image_features(image_batch)
    else:
        return model.get_image_features(image_batch)[0]


@lexy_transformer(name="text.embeddings.clip")
def text_embeddings_clip(
    text: list[str | DocumentBase] | str | DocumentBase,
) -> torch.Tensor:
    """Embed text using CLIP.

    Args:
        text: A single string or DocumentBase instance, or a list of strings or
            DocumentBase instances to embed.

    Returns:
        torch.Tensor: The embeddings of the provided text.
    """
    if isinstance(text, DocumentBase):
        text = text.content
    elif isinstance(text, list):
        text = [s.content if isinstance(s, DocumentBase) else s for s in text]
    tokens = processor(text=text, padding=True, images=None, return_tensors="pt").to(
        device
    )
    text_emb = model.get_text_features(**tokens)
    if isinstance(text, list):
        return text_emb
    else:
        return text_emb[0]



import pytest
from sqlmodel import select

from lexy.models.transformer import Transformer, TransformerCreate


class TestTransformer:
    def test_hello(self):
        assert True

    @pytest.mark.asyncio
    async def test_get_transformers(self, async_session):
        result = await async_session.exec(select(Transformer))
        transformers = result.all()
        assert len(transformers) > 1
        transformer_ids = [t.transformer_id for t in transformers]
        assert "text.embeddings.minilm" in transformer_ids

    def test_get_transformers_with_client(self, client):
        response = client.get(
            "/api/transformers",
        )
        assert response.status_code == 200, response.text
        data = response.json()
        assert len(data) > 1
        transformer_ids = [t["transformer_id"] for t in data]
        assert "text.embeddings.minilm" in transformer_ids

    @pytest.mark.asyncio
    async def test_get_transformers_with_async_client(self, async_client):
        response = await async_client.get(
            "/api/transformers",
        )
        assert response.status_code == 200, response.text
        data = response.json()
        assert len(data) > 1
        transformer_ids = [t["transformer_id"] for t in data]
        assert "text.embeddings.minilm" in transformer_ids

    @pytest.mark.asyncio
    async def test_create_transformer(self, async_session):
        transformer = TransformerCreate(
            transformer_id="test_transformer",
            path="test.tester",
            description="Test Transformer",
        )

        db_transformer = Transformer.model_validate(transformer)
        async_session.add(db_transformer)
        await async_session.commit()
        await async_session.refresh(db_transformer)
        assert db_transformer.transformer_id == "test_transformer"
        assert db_transformer.path == "test.tester"
        assert db_transformer.description == "Test Transformer"
        assert db_transformer.celery_task_name == "lexy.transformers.test_transformer"

        result = await async_session.exec(
            select(Transformer).where(Transformer.transformer_id == "test_transformer")
        )
        transformers = result.all()
        assert len(transformers) == 1
        assert transformers[0].transformer_id == "test_transformer"
        assert transformers[0].path == "test.tester"
        assert transformers[0].description == "Test Transformer"

    @pytest.mark.asyncio
    async def test_delete_transformer(self, async_session):
        result = await async_session.exec(
            select(Transformer).where(Transformer.transformer_id == "test_transformer")
        )
        transformer = result.first()
        assert transformer is not None

        await async_session.delete(transformer)
        await async_session.commit()

        result = await async_session.exec(
            select(Transformer).where(Transformer.transformer_id == "test_transformer")
        )
        transformer = result.first()
        assert transformer is None


class TestTransformerModel:
    def test_create_transformer(self):
        transformer = TransformerCreate(transformer_id="test_transformer")
        assert transformer.transformer_id == "test_transformer"
        transformer = TransformerCreate(transformer_id="test-transformer")
        assert transformer.transformer_id == "test-transformer"
        transformer = TransformerCreate(transformer_id="test.transformer")
        assert transformer.transformer_id == "test.transformer"

    def test_create_transformer_with_invalid_id(self):
        with pytest.raises(ValueError):
            TransformerCreate(
                transformer_id="", description="Test Transformer"
            )  # blank
        with pytest.raises(ValueError):
            TransformerCreate(
                transformer_id="test transformer", description="Test Transformer"
            )  # space
        with pytest.raises(ValueError):
            TransformerCreate(
                transformer_id="_test", description="Test Transformer"
            )  # starts with underscore
        with pytest.raises(ValueError):
            TransformerCreate(
                transformer_id="1abc", description="Test Transformer"
            )  # starts with number
        with pytest.raises(ValueError):
            TransformerCreate(
                transformer_id="transformer" * 30, description="Test Transformer"
            )  # too long



import logging

import boto3
from botocore.exceptions import ClientError, NoCredentialsError

from lexy.storage.base import StorageClient

logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)


class S3Client(StorageClient):
    def __init__(self, **kwargs):
        logger.info("Creating S3 client")
        self.client = boto3.client("s3", **kwargs)

    def is_authenticated(self) -> bool:
        try:
            self.client.list_buckets()
            return True
        except NoCredentialsError:
            return False
        except ClientError:
            return False
        except Exception:
            raise

    def list_buckets(self) -> list[str]:
        response = self.client.list_buckets()
        return [bucket["Name"] for bucket in response["Buckets"]]

    def upload_object(
        self, fileobj, bucket_name: str, object_name: str, rewind: bool = True
    ) -> dict:
        if isinstance(fileobj, str):
            with open(fileobj, "rb") as f:
                fileobj = f
        else:
            if rewind:
                fileobj.seek(0)
        self.client.upload_fileobj(fileobj, bucket_name, object_name)
        return {
            "storage_service": "s3",
            "storage_bucket": bucket_name,
            "storage_key": object_name,
            # "storage_url": f"https://{bucket_name}.s3.amazonaws.com/{object_name}",
            # "storage_uri": f"s3://{bucket_name}/{object_name}",
        }

    def generate_presigned_url(
        self, bucket_name: str, object_name: str, expiration: int = 3600
    ) -> str:
        url = self.client.generate_presigned_url(
            "get_object",
            Params={"Bucket": bucket_name, "Key": object_name},
            ExpiresIn=expiration,
        )
        return url

    def delete_object(self, bucket_name: str, object_name: str) -> None:
        self.client.delete_object(Bucket=bucket_name, Key=object_name)



# source: https://medium.com/cuddle-ai/async-architecture-with-fastapi-celery-and-rabbitmq-c7d029030377
import os

from kombu import Queue


# TODO: fix or delete this
def route_task(name, args, kwargs, options, task=None, **kw):
    if ":" in name:
        queue, _ = name.split(":")
        return {"queue": queue}
    return {"queue": "celery"}


class BaseConfig:
    # TODO: these should be constructed from lexy_settings
    broker_url: str = os.environ.get(
        "CELERY_BROKER_URL", "amqp://guest:guest@localhost:5672//"
    )
    pg_host = os.environ.get("POSTGRES_HOST", "localhost")
    result_backend: str = os.environ.get(
        "CELERY_RESULT_BACKEND",
        f"db+postgresql://postgres:postgres@{pg_host}:5432/lexy",
    )

    broker_connection_retry_on_startup = True
    task_track_started = True

    # Queues
    task_queues: list = (
        # default queue
        Queue("celery", queue_arguments={"x-max-priority": 10}),
        # custom queues
        # Queue("custom_queue_1"),
        # Queue("custom_queue_2"),
    )

    # Default priority settings for all queues
    task_queue_max_priority = 10
    task_default_priority = 5

    # Task routes
    task_routes = {
        "lexy.core.celery_tasks": {"queue": "celery"},
        # 'lexy.transformers.embeddings.text_embeddings': {'queue': 'text:embeddings'},
        # 'lexy.transformers.embeddings.video_embeddings': {'queue': 'video:embeddings'},
    }
    # task_routes = (route_task,)

    # Serialization settings
    task_serializer = "pickle"
    result_serializer = "pickle"
    accept_content = [
        "pickle",
        "json",
        "application/json",
        # 'application/x-python-serialize',  # adding for flower
    ]

    # Result settings
    result_persistent = True
    result_extended = True

    # Worker settings
    # List of modules to import when the Celery worker starts -- these are currently
    # set in `lexy.celery_app.create_celery`.
    # imports = (
    #     'myapp.tasks',
    #     'lexy.core.celery_tasks',
    #     list(lexy_settings.worker_transformer_imports)
    # )
    # include = lexy_settings.worker_transformer_imports
    worker_pool_restarts = True
    worker_send_task_events = True
    worker_prefetch_multiplier = 1
    # source: https://rusty-celery.github.io/best-practices/index.html
    acks_late = True


class DevelopmentConfig(BaseConfig):
    pass


class TestingConfig(BaseConfig):
    # WARNING: These settings will be ignored in favor of any environment variables,
    #  even if set explicitly. See https://github.com/celery/celery/issues/4284.
    #  To override these settings at test time, use `pytest.ini_options` in
    #  pyproject.toml, or overwrite value of env var in lexy_tests/conftest.py.
    broker_url: str = "memory://"
    # TODO: these should be constructed from lexy_settings
    pg_host = os.environ.get("POSTGRES_HOST", "localhost")
    result_backend: str = f"db+postgresql://postgres:postgres@{pg_host}:5432/lexy_tests"


def get_settings():
    config_cls_dict = {
        "development": DevelopmentConfig,
        "testing": TestingConfig,
    }
    config_name = os.environ.get("CELERY_CONFIG", "development")
    config_cls = config_cls_dict[config_name]
    return config_cls()


settings = get_settings()



import os
import warnings

import asyncio
import httpx
import pytest
from fastapi import FastAPI
from fastapi.testclient import TestClient
from sqlalchemy.engine import Engine, make_url
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.session import close_all_sessions
from sqlalchemy.ext.asyncio import async_sessionmaker, create_async_engine
from sqlalchemy.pool import NullPool
from sqlmodel import SQLModel, create_engine, delete
from sqlmodel.ext.asyncio.session import AsyncSession
from asgi_lifespan import LifespanManager

from lexy.core.config import settings, TestAppSettings
from lexy.core.celery_config import settings as celery_settings
from lexy.db.init_db import add_default_data_to_db, add_first_superuser_to_db
from lexy.db.session import get_session
from lexy import models

os.environ["CELERY_BROKER_URL"] = celery_settings.broker_url
os.environ["CELERY_RESULT_BACKEND"] = celery_settings.result_backend

# We need to import the app **after** forcefully setting the env vars, since Celery
# doesn't let you assign any value other than the environment variable if one is set.
# See https://github.com/celery/celery/issues/4284.
from lexy.main import app as lexy_test_app  # noqa: E402


# Values of LEXY_CONFIG and CELERY_CONFIG are set using pytest-env plugin in
# pyproject.toml, but overwritten in lexy_tests/__init__.py
assert os.environ.get("LEXY_CONFIG") == "testing", "LEXY_CONFIG is not set to 'testing'"
assert (
    os.environ.get("CELERY_CONFIG") == "testing"
), "CELERY_CONFIG is not set to 'testing'"


test_settings = settings
test_celery_settings = celery_settings


DB_WARNING_MSG = (
    "There's a good chance you're about to drop the wrong database! "
    "Double check your test settings."
)
assert test_settings.POSTGRES_DB != "lexy", DB_WARNING_MSG


CELERY_DB_WARNING_MSG = (
    "Test instance of Celery is configured to store results in the wrong database! "
    "Double check your test settings."
)
backend_url_obj = make_url(test_celery_settings.result_backend)
assert backend_url_obj.database != "lexy", CELERY_DB_WARNING_MSG


@pytest.fixture(scope="session")
def settings() -> TestAppSettings:
    """Fixture for test settings."""
    print(f"{test_settings = }")
    return test_settings


@pytest.fixture(scope="session")
def sync_engine(settings: TestAppSettings):
    """Create a SQLAlchemy sync engine for the test database."""
    engine = create_engine(
        url=settings.sync_database_url, echo=settings.DB_ECHO_LOG, future=True
    )
    print(f"sync_engine.url: {engine.url}")
    assert engine.url.database != "lexy", DB_WARNING_MSG
    return engine


@pytest.fixture(scope="session")
def async_engine(settings: TestAppSettings):
    """Create a SQLAlchemy async engine for the test database."""
    engine = create_async_engine(
        settings.async_database_url,
        echo=settings.DB_ECHO_LOG,
        future=True,
        poolclass=NullPool,
    )
    print(f"async_engine.url: {engine.url}")
    assert engine.url.database != "lexy", DB_WARNING_MSG
    return engine


@pytest.fixture(scope="session")
def create_test_database(sync_engine, celery_session_app):
    """Create test database and tables."""
    with sync_engine.begin() as conn:
        print(f"Creating test DB tables with engine.url: {sync_engine.url}")
        SQLModel.metadata.create_all(conn)
        print("Created test DB tables")
    yield
    with sync_engine.begin() as conn:
        lexy_table_names = ", ".join(SQLModel.metadata.tables.keys())
        celery_table_names = ", ".join(
            celery_session_app.backend.task_cls.metadata.tables.keys()
        )
        celery_tables = celery_session_app.backend.task_cls.metadata.sorted_tables
        print(
            f"Dropping test DB tables..."
            f"\n\tengine.url: {sync_engine.url}"
            f"\n\tlexy_table_names: {lexy_table_names}"
            f"\n\tcelery_table_names: {celery_table_names}"
        )
        SQLModel.metadata.drop_all(conn)
        celery_session_app.backend.task_cls.metadata.drop_all(
            conn, tables=celery_tables
        )
        print("Dropped test DB tables")


@pytest.fixture(scope="session")
def seed_data(settings: TestAppSettings, sync_engine: Engine, create_test_database):
    """Seed the test database with data."""
    # Create a local sync session for seeding the test database
    local_sessionmaker = sessionmaker(
        autocommit=False, autoflush=False, bind=sync_engine
    )
    local_session = local_sessionmaker()
    # Add seed data to the test database
    print("Seeding the test database with data")
    # Add first superuser
    add_first_superuser_to_db(session=local_session, settings=settings)
    # Add default data
    add_default_data_to_db(session=local_session)
    # Uncomment the next line if you want to add sample documents
    # add_sample_docs_to_db(session=local_session)
    yield
    # Clean up the seed data after the tests
    print("Deleting test DB data...")
    models_to_delete = [
        models.User,
        models.Binding,
        models.Index,
        models.Document,
        models.Collection,
        models.Transformer,
    ]
    for model in models_to_delete:
        # Another way to do it
        # local_session.query(model).delete()
        result = local_session.execute(delete(model))
        deleted_count = result.rowcount
        print(f"\tdeleting {deleted_count} {model.__name__} objects")
    local_session.commit()
    close_all_sessions()
    print("Deleted test DB data")


@pytest.fixture(scope="session")
async def test_app(seed_data) -> FastAPI:
    async with LifespanManager(lexy_test_app):
        yield lexy_test_app


@pytest.fixture(scope="function")
def sync_session(sync_engine, test_app):
    """Create a new sync session for each test case."""
    session = sessionmaker(autocommit=False, autoflush=False, bind=sync_engine)
    with session() as s:
        yield s


@pytest.fixture(scope="function")
async def async_session(async_engine, test_app):
    """Create a new async session for each test case."""
    async_session = async_sessionmaker(
        bind=async_engine, class_=AsyncSession, expire_on_commit=False
    )
    async with async_session() as session:
        yield session


@pytest.fixture(scope="function")
def client(test_app, async_session: AsyncSession) -> TestClient:
    """Fixture for providing a synchronous TestClient configured for testing."""

    async def override_get_session():
        async with async_session as session:
            yield session

    # Override get_session dependency to use the test database session
    test_app.dependency_overrides[get_session] = override_get_session

    client = TestClient(app=test_app, raise_server_exceptions=False)
    yield client

    del test_app.dependency_overrides[get_session]  # Reset overrides after tests


@pytest.fixture(scope="function")
async def async_client(test_app, async_session: AsyncSession) -> httpx.AsyncClient:
    """Fixture for providing an asynchronous TestClient configured for testing."""

    async def override_get_session():
        async with async_session as session:
            yield session

    # Override get_session dependency to use the test database session
    test_app.dependency_overrides[get_session] = override_get_session

    async with httpx.AsyncClient(app=test_app, base_url="http://test") as ac:
        yield ac

    del test_app.dependency_overrides[get_session]  # Reset overrides after tests


@pytest.fixture(scope="function")
def document_storage(settings):
    """Fixture used to skip tests and propagate warnings regarding document storage.

    Tests using this fixture will be skipped if any of the following are true:
        - DEFAULT_STORAGE_BUCKET is not set
        - DEFAULT_STORAGE_SERVICE is 's3' but S3 credentials are unavailable, invalid,
          or lack access
        - DEFAULT_STORAGE_SERVICE is 'gcs' but GOOGLE_APPLICATION_CREDENTIALS is not set
        - DEFAULT_STORAGE_SERVICE is 'gcs' but client is not authenticated
    """
    if settings.DEFAULT_STORAGE_BUCKET is None:
        warnings.warn(
            "DEFAULT_STORAGE_BUCKET is not set - will skip tests requiring document "
            "storage."
        )
        pytest.skip("DEFAULT_STORAGE_BUCKET is not set")

    if settings.DEFAULT_STORAGE_SERVICE == "s3":
        from lexy.storage.s3 import S3Client

        s3_client = S3Client()
        if not s3_client.is_authenticated():
            warnings.warn(
                "DEFAULT_STORAGE_SERVICE is 's3' but credentials are unavailable, "
                "invalid, or lack access - will skip tests requiring document storage."
            )
            pytest.skip("S3 client is not authenticated")

    elif settings.DEFAULT_STORAGE_SERVICE == "gcs":
        if not settings.GOOGLE_APPLICATION_CREDENTIALS:
            warnings.warn(
                "DEFAULT_STORAGE_SERVICE is 'gcs', but GOOGLE_APPLICATION_CREDENTIALS "
                "is not set - will skip tests requiring document storage."
            )
            pytest.skip("GOOGLE_APPLICATION_CREDENTIALS is not set")
        from lexy.storage.gcs import GCSClient

        gcs_client = GCSClient()
        if not gcs_client.is_authenticated():
            warnings.warn(
                "DEFAULT_STORAGE_SERVICE is 'gcs' but credentials are not available - "
                "will skip tests requiring document storage."
            )
            pytest.skip("GCS client is not authenticated")


@pytest.fixture(scope="session")
def celery_settings():
    print(f"{test_celery_settings = }")
    print(f"{test_celery_settings.broker_url = }")
    print(f"{test_celery_settings.result_backend = }")
    return test_celery_settings


@pytest.fixture(scope="session")
def celery_config(settings, celery_settings):
    # These are normally set in lexy.celery_app.create_celery function
    # TODO: add 'lexy.core.celery_tasks' here?
    celery_settings.imports = list(settings.worker_transformer_imports)
    celery_settings.include = settings.worker_transformer_imports

    celery_settings_dict = dict()
    for key in dir(celery_settings):
        if not key.startswith("__"):
            celery_settings_dict[key] = getattr(celery_settings, key)
    print(f"{celery_settings_dict = }")
    return celery_settings_dict


@pytest.fixture(scope="session")
def use_celery_app_trap():
    # Throws an error if a test tries to access the default celery app - can override
    #  in tests that need it using `@pytest.mark.usefixtures('depends_on_current_app')`
    return True


@pytest.fixture(scope="session")
def event_loop():
    loop = asyncio.get_event_loop()
    yield loop
    loop.close()



import logging
import os

from openai import OpenAI

from lexy.models.document import DocumentBase
from lexy.transformers import lexy_transformer


logger = logging.getLogger(__name__)
if os.environ.get("OPENAI_API_KEY"):
    openai_client = OpenAI(api_key=os.environ["OPENAI_API_KEY"])
else:
    openai_client = None
    logger.warning("OPENAI_API_KEY not set; cannot use OpenAI API")


@lexy_transformer(name="text.embeddings.openai")
def text_embeddings(
    text: list[str | DocumentBase] | str | DocumentBase, *, model: str, **kwargs
) -> list[list[float]] | list[float]:
    """Embed text using the OpenAI API.

    Any additional keyword arguments are passed to the client's
    :func:`OpenAI.embeddings.create` method.

    Args:
        text: A single string or DocumentBase instance, or a list of strings or
            DocumentBase instances to embed.
        model: The OpenAI model to use for embeddings. Can be one of
            "text-embedding-ada-002", "text-embedding-3-small", or
            "text-embedding-3-large".

    Keyword Args:
        encoding_format: The format to return the embeddings in. Can be either `float`
            or `base64 <https://pypi.org/project/pybase64/>`__.
        dimensions: The number of dimensions for the resulting output embeddings. Only
            supported in "text-embedding-3" and later models.
        user: A unique identifier representing your end-user, which can help OpenAI to
            monitor and detect abuse.
        extra_headers: Send extra headers
        extra_query: Add additional query parameters to the request
        extra_body: Add additional JSON properties to the request
        timeout: Override the client-level default timeout for this request, in seconds

    Returns:
        list[list[float]] | list[float]: The embeddings of the provided text.
    """
    if not openai_client:
        raise Exception("OPENAI_API_KEY not set; cannot use OpenAI API")

    if isinstance(text, DocumentBase):
        text = text.content
    elif isinstance(text, list):
        text = [s.content if isinstance(s, DocumentBase) else s for s in text]

    api_response = openai_client.embeddings.create(model=model, input=text, **kwargs)

    if isinstance(text, list):
        return [e.embedding for e in api_response.data]
    else:
        return api_response.data[0].embedding


@lexy_transformer(name="text.embeddings.openai-3-small")
def text_embeddings_3_small(
    text: list[str | DocumentBase] | str | DocumentBase, **kwargs
) -> list[list[float]] | list[float]:
    """Embed text using OpenAI's "text-embedding-3-small" model.

    This runs :func:`lexy.transformers.openai.text_embeddings` with
    `model="text-embedding-3-small"`.

    Args:
        text: A single string or DocumentBase instance, or a list of strings or
            DocumentBase instances to embed.

    Returns:
        list[list[float]] | list[float]: The embeddings of the provided text.

    See Also:
        :func:`lexy.transformers.openai.text_embeddings` for additional keyword args.
    """
    return text_embeddings(text, model="text-embedding-3-small", **kwargs)


@lexy_transformer(name="text.embeddings.openai-3-large")
def text_embeddings_3_large(
    text: list[str | DocumentBase] | str | DocumentBase, **kwargs
) -> list[list[float]] | list[float]:
    """Embed text using OpenAI's "text-embedding-3-large" model.

    This runs :func:`lexy.transformers.openai.text_embeddings` with
    `model="text-embedding-3-large"`.

    Args:
        text: A single string or DocumentBase instance, or a list of strings or
            DocumentBase instances to embed.

    Returns:
        list[list[float]] | list[float]: The embeddings of the provided text.

    See Also:
        :func:`lexy.transformers.openai.text_embeddings` for additional keyword args.
    """
    return text_embeddings(text, model="text-embedding-3-large", **kwargs)


@lexy_transformer(name="text.embeddings.openai-ada-002")
def text_embeddings_ada_002(
    text: list[str | DocumentBase] | str | DocumentBase, **kwargs
) -> list[list[float]] | list[float]:
    """Embed text using OpenAI's "text-embedding-ada-002" model.

    This runs :func:`lexy.transformers.openai.text_embeddings` with
    `model="text-embedding-ada-002"`.

    Args:
        text: A single string or DocumentBase instance, or a list of strings or
            DocumentBase instances to embed.

    Returns:
        list[list[float]] | list[float]: The embeddings of the provided text.

    See Also:
        :func:`lexy.transformers.openai.text_embeddings` for additional keyword args.
    """
    return text_embeddings(text, model="text-embedding-ada-002", **kwargs)



import pytest
from sqlmodel import select

from lexy.models.collection import Collection, CollectionCreate, CollectionUpdate


class TestCollection:
    def test_hello(self):
        assert True

    @pytest.mark.asyncio
    async def test_get_collections(self, async_session):
        result = await async_session.exec(select(Collection))
        collections = result.all()
        assert len(collections) > 1
        collection_names = [c.collection_name for c in collections]
        assert "default" in collection_names

    def test_get_collections_with_client(self, client):
        response = client.get(
            "/api/collections",
        )
        assert response.status_code == 200, response.text
        data = response.json()
        assert len(data) > 1
        collection_names = [c["collection_name"] for c in data]
        assert "default" in collection_names

    @pytest.mark.asyncio
    async def test_get_collections_with_async_client(self, async_client):
        response = await async_client.get(
            "/api/collections",
        )
        assert response.status_code == 200, response.text
        data = response.json()
        assert len(data) > 1
        collection_names = [c["collection_name"] for c in data]
        assert "default" in collection_names

    @pytest.mark.asyncio
    async def test_create_collection(self, async_session, settings):
        collection = Collection(
            collection_name="test_collection", description="Test Collection"
        )
        async_session.add(collection)
        await async_session.commit()
        await async_session.refresh(collection)
        assert collection.collection_id is not None
        assert collection.collection_name == "test_collection"
        assert collection.description == "Test Collection"
        assert collection.config == settings.COLLECTION_DEFAULT_CONFIG
        assert collection.created_at is not None
        assert collection.updated_at is not None

        result = await async_session.exec(
            select(Collection).where(Collection.collection_name == "test_collection")
        )
        collections = result.all()
        assert len(collections) == 1
        assert collections[0].collection_name == "test_collection"
        assert collections[0].description == "Test Collection"
        assert collections[0].config == settings.COLLECTION_DEFAULT_CONFIG

    @pytest.mark.asyncio
    async def test_create_collection_with_model_validate(self, async_session, settings):
        collection = CollectionCreate(
            collection_name="test_collection_validated",
            description="Test Collection Validated",
        )
        assert collection.collection_name == "test_collection_validated"
        assert collection.description == "Test Collection Validated"
        assert collection.config == settings.COLLECTION_DEFAULT_CONFIG

        db_collection = Collection.model_validate(collection)
        async_session.add(db_collection)
        await async_session.commit()
        await async_session.refresh(db_collection)
        assert db_collection.collection_id is not None
        assert db_collection.collection_name == "test_collection_validated"
        assert db_collection.description == "Test Collection Validated"
        assert db_collection.config == settings.COLLECTION_DEFAULT_CONFIG
        assert db_collection.created_at is not None
        assert db_collection.updated_at is not None

        result = await async_session.exec(
            select(Collection).where(
                Collection.collection_name == "test_collection_validated"
            )
        )
        collections = result.all()
        assert len(collections) == 1
        assert collections[0].collection_name == "test_collection_validated"
        assert collections[0].description == "Test Collection Validated"
        assert collections[0].config == settings.COLLECTION_DEFAULT_CONFIG

    @pytest.mark.asyncio
    async def test_collection_crud(self, async_session):
        # create collection
        collection = CollectionCreate(
            collection_name="test_collection_crud", description="Test Collection CRUD"
        )
        db_collection = Collection.model_validate(collection)
        async_session.add(db_collection)
        await async_session.commit()
        await async_session.refresh(db_collection)
        assert db_collection.collection_id is not None
        assert db_collection.collection_name == "test_collection_crud"
        assert db_collection.description == "Test Collection CRUD"
        assert db_collection.created_at is not None
        assert db_collection.updated_at is not None

        # get collection
        result = await async_session.exec(
            select(Collection).where(
                Collection.collection_name == "test_collection_crud"
            )
        )
        collections = result.all()
        assert len(collections) == 1
        assert collections[0].collection_name == "test_collection_crud"
        assert collections[0].description == "Test Collection CRUD"

        # update collection
        collection_update = CollectionUpdate(description="Test Collection CRUD Updated")
        collection_data = collection_update.model_dump(exclude_unset=True)
        for key, value in collection_data.items():
            setattr(db_collection, key, value)
        async_session.add(db_collection)
        await async_session.commit()
        await async_session.refresh(db_collection)
        assert db_collection.collection_id is not None
        assert db_collection.collection_name == "test_collection_crud"
        assert db_collection.description == "Test Collection CRUD Updated"
        assert db_collection.created_at is not None
        assert db_collection.updated_at > db_collection.created_at

        # delete collection
        result = await async_session.exec(
            select(Collection).where(
                Collection.collection_name == "test_collection_crud"
            )
        )
        collection = result.first()
        assert collection is not None
        await async_session.delete(collection)
        await async_session.commit()

        result = await async_session.exec(
            select(Collection).where(
                Collection.collection_name == "test_collection_crud"
            )
        )
        collection = result.first()
        assert collection is None


class TestCollectionModel:
    def test_create_collection_model(self):
        collection = CollectionCreate(collection_name="test_collection")
        assert collection.collection_name == "test_collection"
        collection = CollectionCreate(collection_name="_te5t")
        assert collection.collection_name == "_te5t"
        collection = CollectionCreate(collection_name="_mytable" * 7)
        assert collection.collection_name == "_mytable" * 7

    def test_create_collection_model_with_invalid_name(self):
        with pytest.raises(ValueError):
            CollectionCreate(collection_name="", description="Test Collection")  # blank
        with pytest.raises(ValueError):
            CollectionCreate(
                collection_name="test collection", description="Test Collection"
            )  # space
        with pytest.raises(ValueError):
            CollectionCreate(
                collection_name="test-collection", description="Test Collection"
            )  # hyphen
        with pytest.raises(ValueError):
            CollectionCreate(
                collection_name="Test", description="Test Collection"
            )  # uppercase
        with pytest.raises(ValueError):
            CollectionCreate(
                collection_name="1abc", description="Test Collection"
            )  # starts with number
        with pytest.raises(ValueError):
            CollectionCreate(
                collection_name="_mytable" * 8, description="Test Collection"
            )  # too long

    def test_update_collection_model(self):
        collection_update = CollectionUpdate(collection_name="test_collection")
        assert collection_update.collection_name == "test_collection"
        collection_update = CollectionUpdate(
            collection_name="_te5t", description="Updated description"
        )
        assert collection_update.collection_name == "_te5t"
        assert collection_update.description == "Updated description"
        collection_update = CollectionUpdate(description="Updated description")
        assert collection_update.collection_name is None
        assert collection_update.description == "Updated description"

    def test_update_collection_model_with_invalid_name(self):
        with pytest.raises(ValueError):
            CollectionUpdate(collection_name="", description="Test Collection")  # blank
        with pytest.raises(ValueError):
            CollectionUpdate(
                collection_name="test collection", description="Test Collection"
            )  # space
        with pytest.raises(ValueError):
            CollectionUpdate(
                collection_name="test-collection", description="Test Collection"
            )  # hyphen
        with pytest.raises(ValueError):
            CollectionUpdate(
                collection_name="Test", description="Test Collection"
            )  # uppercase
        with pytest.raises(ValueError):
            CollectionUpdate(
                collection_name="1abc", description="Test Collection"
            )  # starts with number
        with pytest.raises(ValueError):
            CollectionUpdate(
                collection_name="_mytable" * 8, description="Test Collection"
            )  # too long



import logging
from typing import Optional, TYPE_CHECKING

from fastapi import APIRouter, Depends, HTTPException, status
from fastapi.encoders import jsonable_encoder
from sqlmodel import select
from sqlmodel.ext.asyncio.session import AsyncSession

from lexy import crud
from lexy.db.session import get_session
from lexy.storage.client import get_storage_client
from lexy.models.binding import Binding, BindingCreate, BindingRead, BindingUpdate
from lexy.core.events import process_new_binding

if TYPE_CHECKING:
    from lexy.storage.client import StorageClient


logger = logging.getLogger(__name__)
logger.setLevel(logging.INFO)
router = APIRouter()


@router.get(
    "/bindings",
    response_model=list[BindingRead],
    status_code=status.HTTP_200_OK,
    name="get_bindings",
    description="Get all bindings",
)
async def get_bindings(
    session: AsyncSession = Depends(get_session),
) -> list[BindingRead]:
    result = await session.exec(select(Binding))
    bindings = result.all()
    return bindings


# TODO: change to the following after SQLAlchemy 2.0: https://stackoverflow.com/a/75947988
@router.post(
    "/bindings",
    response_model=dict[str, BindingRead | list[dict]],
    status_code=status.HTTP_201_CREATED,
    name="create_binding",
    description="Create a new binding",
)
async def create_binding(
    binding: BindingCreate,
    session: AsyncSession = Depends(get_session),
    storage_client: Optional["StorageClient"] = Depends(get_storage_client),
) -> dict[str, BindingRead | list[dict]]:
    # set collection_id based on collection_id or collection_name
    if binding.collection_id:
        collection = await crud.get_collection_by_id(
            session=session, collection_id=binding.collection_id
        )
    else:
        # BindingCreate model validator ensures that either id or name is provided
        collection = await crud.get_collection_by_name(
            session=session, collection_name=binding.collection_name
        )
    if not collection:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST, detail="Collection not found"
        )
    binding.collection_id = collection.collection_id

    transformer = await crud.get_transformer_by_id(
        session=session, transformer_id=binding.transformer_id
    )
    if not transformer:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST, detail="Transformer not found"
        )
    index = await crud.get_index_by_id(session=session, index_id=binding.index_id)
    if not index:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST, detail="Index not found"
        )

    # TODO: switch to pattern `db_binding = Binding.model_validate(binding)` once issue
    #  is resolved. Currently SQLModel is not serializing the nested model, leading to
    #  the error 'Filter is not JSON Serializable'. But using
    #  `Binding(**binding.model_dump())` leaves binding.filter as a dict type.
    #  See issue: https://github.com/tiangolo/sqlmodel/issues/63
    db_binding = Binding(**binding.model_dump())
    session.add(db_binding)
    await session.commit()
    await session.refresh(db_binding)

    # process the binding and generate document tasks
    processed_binding, tasks = await session.run_sync(
        process_new_binding, db_binding, storage_client=storage_client
    )
    # now commit the binding again and refresh it - status should be updated
    session.add(processed_binding)
    await session.commit()
    await session.refresh(processed_binding)

    response = {"binding": processed_binding, "tasks": tasks}
    return response


@router.get(
    "/bindings/{binding_id}",
    response_model=BindingRead,
    status_code=status.HTTP_200_OK,
    name="get_binding",
    description="Get a binding",
)
async def get_binding(
    binding_id: int, session: AsyncSession = Depends(get_session)
) -> BindingRead:
    result = await session.exec(select(Binding).where(Binding.binding_id == binding_id))
    binding = result.first()
    if not binding:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Binding not found"
        )
    return binding


# TODO: change to the following after SQLAlchemy 2.0: https://stackoverflow.com/a/75947988
@router.patch(
    "/bindings/{binding_id}",
    status_code=status.HTTP_200_OK,
    name="update_binding",
    description="Update a binding",
)
async def update_binding(
    binding_id: int,
    binding: BindingUpdate,
    session: AsyncSession = Depends(get_session),
    storage_client: Optional["StorageClient"] = Depends(get_storage_client),
) -> dict[str, BindingRead | list[dict]]:
    result = await session.exec(select(Binding).where(Binding.binding_id == binding_id))
    db_binding = result.first()
    if not db_binding:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Binding not found"
        )
    # record status in case it's changed
    old_status = db_binding.status

    update_data = binding.model_dump(exclude_unset=True)
    for key, value in update_data.items():
        setattr(db_binding, key, value)
    if db_binding.filter:
        db_binding.filter = jsonable_encoder(db_binding.filter)
    session.add(db_binding)
    await session.commit()
    await session.refresh(db_binding)

    # if status has changed, process the binding
    if db_binding.status == "on" and old_status != "on":
        # TODO: this portion needs to be updated to reflect time stamps of tasks
        #  or to simply become part of an init script
        print(
            f"Binding status changed from '{old_status}' to 'on' - processing binding"
        )
        processed_binding, tasks = await session.run_sync(
            process_new_binding, db_binding, storage_client=storage_client
        )
        # now commit the binding again and refresh it - status should be updated
        session.add(processed_binding)
        await session.commit()
        await session.refresh(processed_binding)
        response = {"binding": processed_binding, "tasks": tasks}
    else:
        response = {"binding": db_binding, "tasks": []}
    return response


@router.delete(
    "/bindings/{binding_id}",
    status_code=status.HTTP_200_OK,
    name="delete_binding",
    description="Delete a binding",
)
async def delete_binding(
    binding_id: int, session: AsyncSession = Depends(get_session)
) -> dict:
    result = await session.exec(select(Binding).where(Binding.binding_id == binding_id))
    binding = result.first()
    if not binding:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail="Binding not found"
        )
    await session.delete(binding)
    await session.commit()
    return {"msg": "Binding deleted", "binding_id": binding_id}



import base64
from datetime import datetime
from io import BytesIO
from typing import Any, Optional, TYPE_CHECKING
from uuid import uuid4, UUID

import httpx
from PIL import Image
from pydantic import PrivateAttr
from sqlalchemy import Column, DateTime, func
from sqlalchemy.dialects.postgresql import JSONB
from sqlmodel import Field, SQLModel, Relationship

from lexy.models.collection import Collection
from lexy.storage import presigned_url_is_expired
from lexy.storage.client import generate_signed_urls_for_document

if TYPE_CHECKING:
    from lexy.storage.client import StorageClient


class DocumentBase(SQLModel):
    content: str
    meta: Optional[dict[Any, Any]] = Field(sa_column=Column(JSONB), default={})
    _image: Optional["Image"] = PrivateAttr(default=None)

    # TODO: this __init__ is only needed because SQLModel is missing
    #  `_init_private_attributes` - should no longer be needed after updating to
    #  Pydantic v2. Sources:
    #  - https://github.com/tiangolo/sqlmodel/pull/472#issuecomment-1301647570
    #  - https://github.com/tiangolo/sqlmodel/issues/504
    def __init__(self, **data: Any):
        super().__init__(**data)
        self._image = None

    @property
    def is_stored_object(self) -> bool:
        return bool(
            self.meta.get("storage_service")
            and self.meta.get("storage_bucket")
            and self.meta.get("storage_key")
        )

    @property
    def image(self) -> Image:
        if not self._image:
            self._image = self.meta.get("image", {}).get("im")
        if not self._image:
            base64_str = self.meta.get("image", {}).get("base64")
            if base64_str:
                self._image = self.image_from_base64_str(base64_str)
        if not self._image:
            if self.object_url:
                self._image = self.image_from_url(self.object_url)
        return self._image

    @property
    def object_url(self) -> Optional[str]:
        if not self.meta.get("_urls"):
            self.refresh_object_urls()
        url = self.meta.get("_urls", {}).get("object")
        # check if url is expired and refresh if needed
        storage_service = self.meta.get("storage_service")
        if url and presigned_url_is_expired(url, storage_service=storage_service):
            self.refresh_object_urls()
            url = self.meta.get("_urls", {}).get("object")
        return url

    # FIXME: storage_client is required, but not passed in with object_url's
    #  self.refresh_object_urls()
    def refresh_object_urls(
        self, storage_client: "StorageClient" = None, expiration: int = 3600
    ) -> None:
        urls = generate_signed_urls_for_document(
            self, storage_client=storage_client, expiration=expiration
        )
        self.meta["_urls"] = urls

    # TODO: move to future ImageDocument class
    @staticmethod
    def image_from_base64_str(base64_str: str) -> Image:
        img_bytes = base64.b64decode(base64_str)
        return Image.open(BytesIO(img_bytes))

    # TODO: move to future ImageDocument class
    @staticmethod
    def image_from_url(url: str) -> Image:
        r = httpx.get(url)
        return Image.open(BytesIO(r.content))


class Document(DocumentBase, table=True):
    __tablename__ = "documents"
    document_id: UUID = Field(
        default_factory=uuid4,
        primary_key=True,
        index=True,
        nullable=False,
    )
    created_at: datetime = Field(
        default=None,
        sa_column=Column(
            DateTime(timezone=True), nullable=False, server_default=func.now()
        ),
    )
    updated_at: datetime = Field(
        default=None,
        sa_column=Column(
            DateTime(timezone=True),
            nullable=False,
            server_default=func.now(),
            onupdate=func.now(),
        ),
    )
    collection_id: str = Field(default=None, foreign_key="collections.collection_id")
    collection: Collection = Relationship(
        back_populates="documents", sa_relationship_kwargs={"lazy": "selectin"}
    )


class DocumentCreate(DocumentBase):
    document_id: Optional[UUID] = None


class DocumentUpdate(DocumentBase):
    content: Optional[str] = None
    meta: Optional[dict[Any, Any]] = None



import pytest
from sqlmodel import select

from lexy import crud
from lexy.models.binding import Binding, BindingCreate, BindingRead, BindingUpdate
from lexy.models.collection import Collection
from lexy.models.index import Index
from lexy.models.transformer import Transformer


class TestBinding:
    def test_hello(self):
        assert True

    @pytest.mark.asyncio
    async def test_get_bindings(self, async_session):
        result = await async_session.exec(select(Binding))
        bindings = result.all()
        assert len(bindings) > 0

        b = bindings[0]
        assert b.binding_id == 1
        assert b.description == "Default binding"

        # these should be present for BindingRead class only
        assert "collection" not in b.model_dump()
        assert "transformer" not in b.model_dump()
        assert "index" not in b.model_dump()

    def test_get_bindings_with_client(self, client):
        response = client.get(
            "/api/bindings",
        )
        assert response.status_code == 200, response.text
        data = response.json()
        assert len(data) > 0

        b = data[0]
        assert b["binding_id"] == 1
        assert b["description"] == "Default binding"

        # these should be present for BindingRead class only
        assert "collection" in b
        assert "transformer" in b
        assert "index" in b

        binding = BindingRead.model_validate(b)
        assert binding.binding_id == 1
        assert isinstance(binding.collection, Collection)
        assert binding.collection.collection_name == "default"
        assert isinstance(binding.index, Index)
        assert binding.index.index_id == "default_text_embeddings"
        assert isinstance(binding.transformer, Transformer)
        assert binding.transformer.transformer_id == "text.embeddings.minilm"

    @pytest.mark.asyncio
    async def test_get_bindings_with_async_client(self, async_client):
        response = await async_client.get(
            "/api/bindings",
        )
        assert response.status_code == 200, response.text
        data = response.json()
        assert len(data) > 0

        b = data[0]
        assert b["binding_id"] == 1
        assert b["description"] == "Default binding"

        # these should be present for BindingRead class only
        assert "collection" in b
        assert "transformer" in b
        assert "index" in b

        binding = BindingRead.model_validate(b)
        assert binding.binding_id == 1
        assert isinstance(binding.collection, Collection)
        assert binding.collection.collection_name == "default"
        assert isinstance(binding.index, Index)
        assert binding.index.index_id == "default_text_embeddings"
        assert isinstance(binding.transformer, Transformer)
        assert binding.transformer.transformer_id == "text.embeddings.minilm"

    @pytest.mark.asyncio
    async def test_create_and_delete_binding(self, async_session):
        b_dict = {
            "collection_name": "default",
            "transformer_id": "text.embeddings.minilm",
            "index_id": "default_text_embeddings",
            "description": "Test binding with filter",
            "execution_params": {},
            "transformer_params": {},
            "filter": {
                "conditions": [
                    {
                        "field": "meta.size",
                        "operation": "less_than",
                        "value": 30000.0,
                        "negate": False,
                    }
                ],
                "combination": "AND",
            },
        }
        b = BindingCreate.model_validate(b_dict)
        collection = await crud.get_collection_by_name(
            session=async_session, collection_name=b.collection_name
        )
        assert collection is not None
        assert collection.collection_id is not None
        assert collection.collection_name == "default"
        b.collection_id = collection.collection_id

        binding = Binding(**b.model_dump())
        async_session.add(binding)
        await async_session.commit()
        await async_session.refresh(binding)
        assert binding.binding_id is not None
        assert binding.description == "Test binding with filter"
        assert binding.created_at is not None
        assert binding.updated_at is not None
        assert binding.collection_id == collection.collection_id

        result = await async_session.exec(
            select(Binding).where(Binding.binding_id == binding.binding_id)
        )
        bindings = result.all()
        assert len(bindings) == 1
        assert bindings[0].binding_id == binding.binding_id
        assert bindings[0].description == "Test binding with filter"

        # delete binding
        await async_session.delete(bindings[0])
        await async_session.commit()

        result = await async_session.exec(
            select(Binding).where(Binding.binding_id == binding.binding_id)
        )
        bindings = result.all()
        assert len(bindings) == 0


class TestBindingModel:
    def test_create_binding(self):
        binding = BindingCreate(
            collection_name="default",
            index_id="default_text_embeddings",
            transformer_id="text.embeddings.minilm",
        )
        assert binding.collection_name == "default"
        assert binding.index_id == "default_text_embeddings"
        assert binding.transformer_id == "text.embeddings.minilm"

    def test_create_binding_with_invalid_identifiers(self):
        with pytest.raises(ValueError):
            # collection identifiers missing
            BindingCreate(
                collection_id=None,
                collection_name=None,
                transformer_id="tid",
                index_id="iid",
                description="Binding with no valid collection identifier",
            )
        with pytest.raises(ValueError):
            # transformer identifiers missing
            BindingCreate(
                collection_id="cid",
                transformer_id=None,
                index_id="iid",
                description="Binding with no valid transformer identifier",
            )
        with pytest.raises(ValueError):
            # index identifiers missing
            BindingCreate(
                collection_id="cid",
                transformer_id="tid",
                description="Binding with no valid index identifier",
            )

    def test_update_binding(self):
        binding_update = BindingUpdate(
            description="Updated description",
            filter={
                "conditions": [
                    {
                        "field": "meta.size",
                        "operation": "less_than",
                        "value": 30000,
                        "negate": False,
                    }
                ],
                "combination": "AND",
            },
        )
        assert binding_update.description == "Updated description"
        assert binding_update.filter is not None
        assert binding_update.execution_params is None
        assert binding_update.transformer_params is None
        assert binding_update.status is None



import io

import sqlalchemy as sa
from fastapi import (
    APIRouter,
    Depends,
    File,
    Form,
    HTTPException,
    Query,
    UploadFile,
    status,
)

from PIL import Image
from sqlalchemy import asc, func
from sqlalchemy.orm import aliased
from sqlmodel import SQLModel, select
from sqlmodel.ext.asyncio.session import AsyncSession

from lexy.core.events import celery
from lexy.core.celery_tasks import convert_arrays_to_lists
from lexy.db.session import get_session
from lexy.models.document import Document
from lexy.models.index import Index
from lexy.models.transformer import Transformer
from lexy.api.deps import index_manager


router = APIRouter()

document_tbl = aliased(Document, name="document_tbl")


@router.get(
    "/indexes/{index_id}/records",
    response_model=list[dict],
    status_code=status.HTTP_200_OK,
    name="get_records",
    description="Get records for an index",
)
async def get_records(
    index_id: str = "default_text_embeddings",
    document_id: str | None = None,
    session: AsyncSession = Depends(get_session),
) -> list[dict]:
    result = await session.exec(select(Index).where(Index.index_id == index_id))
    index = result.first()
    if not index:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND, detail=f"Index {index_id} not found"
        )

    index_model = index_manager.index_models.get(index.index_id)
    statement = select(index_model)
    if document_id:
        statement = statement.where(index_model.document_id == document_id)
    result = await session.exec(statement)
    index_records = result.all()
    # FIXME: need a more efficient way to do this - revisit after upgrading to Pydantic 2.x
    return [convert_arrays_to_lists(ir.model_dump()) for ir in index_records]


@router.post(
    "/indexes/{index_id}/records/query",
    response_model=dict,
    status_code=status.HTTP_200_OK,
    name="query_records",
    description="Query all records for an index",
)
async def query_records(
    query_text: str = Form(None),
    query_image: UploadFile = File(None),
    k: int = 5,
    query_field: str = "embedding",
    index_id: str = "default_text_embeddings",
    return_fields: list[str] = Query(None),
    return_document: bool = False,
    embedding_model: str = None,
    session: AsyncSession = Depends(get_session),
) -> dict:
    if query_text and query_image:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Please submit either 'query_text' or 'query_image', not both.",
        )
    elif query_text:
        query = query_text
    elif query_image:
        file_content = await query_image.read()
        query = Image.open(io.BytesIO(file_content))
    else:
        raise HTTPException(
            status_code=status.HTTP_400_BAD_REQUEST,
            detail="Please submit either 'query_text' or 'query_image'.",
        )

    # get index table and query column
    result = await session.exec(select(Index).where(Index.index_id == index_id))
    index = result.first()
    if not index:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Index '{index_id}' not found",
        )
    index_tbl = SQLModel.metadata.tables.get(index.index_table_name)
    query_column = index_tbl.c.get(query_field, None)
    if query_column is None:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Field '{query_field}' not found in index '{index_id}'",
        )

    # get embedding model using index_fields
    if embedding_model is None:
        try:
            embedding_model: str = index.index_fields[query_field]["extras"]["model"]
        except KeyError:
            raise HTTPException(
                status_code=status.HTTP_400_BAD_REQUEST,
                detail=f"Index field '{query_field}' does not have an embedding model. "
                f"You can specify one using the 'embedding_model' parameter.",
            )
    # if multimodal, assign model based on query type
    if "*" in embedding_model:
        if isinstance(query, Image.Image):
            embedding_model = embedding_model.replace("*", "image")
        else:
            embedding_model = embedding_model.replace("*", "text")

    # get embedding for query string
    transformer = await session.exec(
        select(Transformer).where(Transformer.transformer_id == embedding_model)
    )
    transformer = transformer.first()
    if not transformer:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Transformer '{embedding_model}' not found",
        )
    task = celery.send_task(transformer.celery_task_name, args=[query], priority=10)
    result = task.get()
    if isinstance(result, list):
        query_embedding = result
    else:
        query_embedding = result.tolist()

    # get index fields to return
    if return_fields:
        return_index_fields = []
        # get any related document fields
        for rf in return_fields:
            if rf.startswith("document."):
                attribute_name = rf.split(".")[1]
                # if document field doesn't exist, return an error
                if attribute_name not in document_tbl.__table__.columns.keys():
                    raise HTTPException(
                        status_code=status.HTTP_400_BAD_REQUEST,
                        detail=f"Field '{attribute_name}' not found in document",
                    )
                return_index_fields.append(
                    getattr(document_tbl, attribute_name).label(rf)
                )
            else:
                # if index field doesn't exist, return an error
                if rf not in index_tbl.c.keys():
                    raise HTTPException(
                        status_code=status.HTTP_400_BAD_REQUEST,
                        detail=f"Field '{rf}' not found in index '{index_id}'",
                    )
                return_index_fields.append(index_tbl.c[rf])
    else:
        return_index_fields = [
            index_tbl.c[k]
            for k, v in index.index_fields.items()
            if v["type"] != "embedding"
        ]

    # if there's nothing else to return, return document content
    # TODO: remove this condition once we have auto-embed option which will return the
    #  embedded content
    if return_document is False and len(return_index_fields) == 0:
        return_index_fields.append(document_tbl.content.label("document.content"))

    base_query = sa.select(
        index_tbl.c.document_id,
        index_tbl.c.custom_id,
        index_tbl.c.meta,
        index_tbl.c.index_record_id,
        func.abs(query_column.op("<->")(query_embedding)).label("abs_distance"),
        # not adding a function here causes the query to fail for some reason
        func.pow(query_column.op("<->")(query_embedding), 1).label("distance"),
        *return_index_fields,
    ).join(document_tbl, index_tbl.c.document_id == document_tbl.document_id)

    # optionally return the document object
    # using this as a hack to grab document fields - will refactor to use SQLModel
    # objects instead of tables
    doc_prefix = "doc_"
    if return_document:
        document_columns = [
            getattr(document_tbl, field).label(doc_prefix + field)
            for field in document_tbl.__table__.columns.keys()
        ]
        base_query = base_query.add_columns(*document_columns)

    # query index table
    search_result = await session.exec(base_query.order_by(asc("distance")).limit(k))
    search_results = search_result.all()

    # process results to nest document fields under "document" key - will get rid of
    # this once we use SQLModel objects
    formatted_results = []
    for row in search_results:
        result_dict = row._asdict()  # Convert result row to dictionary
        document_data = {}
        if return_document:
            # Extract prefixed Document fields and nest them under 'document' key
            for field in document_tbl.__table__.columns.keys():
                prefixed_name = doc_prefix + field
                document_data[field] = result_dict.pop(prefixed_name)
            result_dict["document"] = document_data
        formatted_results.append(result_dict)

    return {"search_results": formatted_results}


@router.get(
    "/indexes/{index_id}/records/{index_record_id}",
    status_code=status.HTTP_200_OK,
    name="get_record",
    description="Get a record from an index",
)
async def get_record(
    index_record_id: str, index_id: str, session: AsyncSession = Depends(get_session)
) -> dict:
    index_result = await session.exec(select(Index).where(Index.index_id == index_id))
    index = index_result.first()
    if not index:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Index '{index_id}' not found",
        )
    index_model = index_manager.index_models.get(index.index_id)
    result = await session.exec(
        select(index_model).where(index_model.index_record_id == index_record_id)
    )
    index_record = result.first()
    if not index_record:
        raise HTTPException(
            status_code=status.HTTP_404_NOT_FOUND,
            detail=f"Index record '{index_record_id}' not found in index '{index_id}'",
        )
    # FIXME: need a more efficient way to do this - revisit after upgrading to Pydantic 2.x
    return convert_arrays_to_lists(index_record.model_dump())
